import importlib
import inspect
import logging
import os
import traceback
import sys

logger = logging.getLogger(__name__)


def discover_patches():
    """
        Attempt to load patches from the installed patches dir as well as any patches directory in the current dir.
    """

    patches = discover_patches_from_dir("troi.patches.", os.path.join(os.path.dirname(__file__), "patches"))
    try:
        local_patches = discover_patches_from_dir("patches.", "./patches", True)
    except FileNotFoundError:
        local_patches = {}

    return  {**patches, **local_patches}


def discover_patches_from_dir(module_path, patch_dir, add_dot=False):
    """
        Load patches given the appropriate python module path and then file system path. 
        If add_dot = True, add . to the sys.path and then remove it before this function exists.
    """

    # This is here to avoid a circular import
    import troi.patch
    if add_dot:
        sys.path.append(".")

    patch_dict = {}
    for path in os.listdir(patch_dir):
        if path in ['.', '..']:
            continue

        if os.path.isdir(os.path.join(patch_dir, path)):
            continue

        if path.endswith(".py"):
            try:
                patch = importlib.import_module(module_path + path[:-3])
            except ImportError as err:
                logger.info("Cannot import %s, skipping:" % (path), file=sys.stderr)
                traceback.print_exc()
                continue

            for member in inspect.getmembers(patch):
                if inspect.isclass(member[1]):
                    if issubclass(member[1], troi.patch.Patch):
                        if member[1].slug() is not None:
                            patch_dict[member[1].slug()] = member[1]

    if add_dot:
        sys.path.pop(-1)

    return patch_dict


def recursively_update_dict(source, overrides):
    """ Updates the `source` dictionary in place and in a recursive fashion. That is unlike
    dict1.update(dict2) which would simply replace values of keys even in case one of the
    values is dict, this method will attempt to merge the nested dicts.

    Eg: dict1 - {"a": {"b": 1}}, dict2 - {"a": {"c": 2}}
    dict1.update(dict2) - {"a": {"c": 2}}
    recursively_update_dict(dict1, dict2) - {"a": {"b": 1, "c": 2}}
    """
    for key, value in overrides.items():
        if isinstance(value, dict) and value:
            source[key] = recursively_update_dict(source.get(key, {}), value)
        else:
            source[key] = overrides[key]
    return source


def interleave(lists):
    """ Return a list with all items from the given lists."""

    result = []

    while True:
        added = False
        for l in lists:
            try:
                result.append(l.pop(0))
                added = True
            except IndexError:
                pass

        if not added:
            break

    return result




import unittest

from troi.plist import plist


class TestSplitter(unittest.TestCase):

    def test_plist(self):
        pl = plist([0,1,2,3,4,5,6,7,8,9])

        assert pl[0:10] == [0]
        assert pl[0:20] == [0,1]
        assert pl[50:100] == [5,6,7,8,9]
        assert pl.uslice(0, 10) == [0]
        assert pl.uslice(0, 20) == [0,1]
        assert pl.uslice(50, 100) == [5,6,7,8,9]

        assert pl.dslice(0, 2) == [0,1]

        assert pl.random_item(50, 100) in [5,6,7,8,9]

    def test_plist_unique(self):
        pl = plist([0,1,2,3,4,5,6,7,8,9])
        rlist = pl.random_item(count=9)
        assert len(rlist) == len(set(rlist))



import logging

from troi.content_resolver.model.recording import Recording, FileIdType
from troi.content_resolver.unresolved_recording import UnresolvedRecordingTracker
from troi.content_resolver.fuzzy_index import FuzzyIndex
from lb_matching_tools.cleaner import MetadataCleaner
from troi.content_resolver.utils import bcolors

logger = logging.getLogger(__name__)


class ContentResolver:
    '''
    Scan a given path and enter/update the metadata in the search index
    '''

    def __init__(self, quiet):
        self.fuzzy_index = None
        self.quiet = quiet

    def get_artist_recording_metadata(self):
        """
            Fetch the metadata needed to build a fuzzy search index.
        """

        artist_recording_data = []
        for recording in Recording.select():
            artist_recording_data.append((recording.artist_name, recording.recording_name, recording.id))

        return artist_recording_data

    def build_index(self):
        """
            Fetch the data from the DB and then build the fuzzy lookup index.
        """

        artist_recording_data = self.get_artist_recording_metadata()
        for recording in Recording.select():
            artist_recording_data.append((recording.artist_name, recording.recording_name, recording.id))

        self.fuzzy_index = FuzzyIndex()
        self.fuzzy_index.build(artist_recording_data)

    def resolve_recordings(self, query_data, match_threshold):
        """
        Given a list of dicts with artist_name, recording_name, recording_mbid in query data and
        a matching threshold, attempt to match recordings by looking them up in the fuzzy index.
        """

        resolved_recordings = []
        unresolved_recording_mbids = []

        # Set indexes in the data so we can correlate matches
        for i, data in enumerate(query_data):
            data["index"] = i

        mc = MetadataCleaner()
        while True:
            next_query_data = []
            hits = self.fuzzy_index.search(query_data)
            for hit, data in zip(hits, query_data):

                # If we resolved this recording via MBID in a previous step, accept that as a match
                if "recording_id" in data:
                    resolved_recordings.append({
                        "artist_name": data["artist_name"],
                        "recording_name": data["recording_name"],
                        "recording_mbid": data["recording_mbid"],
                        "recording_id": data["recording_id"],
                        "confidence": 1.0,
                        "index": data["index"],
                        "method": "MBID"
                    })
                    continue

                # if not, proceed to examine the match that metadata lookup gave us
                if hit["confidence"] < match_threshold:
                    next_query_data.append(data)
                    unresolved_recording_mbids.append(data["recording_mbid"])
                else:
                    resolved_recordings.append({
                        "artist_name": data["artist_name"],
                        "recording_name": data["recording_name"],
                        "recording_mbid": data["recording_mbid"],
                        "recording_id": hit["recording_id"],
                        "confidence": hit["confidence"],
                        "index": data["index"],
                        "method": "FUZZY"
                    })

            if len(next_query_data) == 0:
                break

            query_data = []
            for data in next_query_data:
                recording_name = mc.clean_recording(data["recording_name"])
                artist_name = mc.clean_artist(data["artist_name"])
                if recording_name != data["recording_name"]:
                    query_data.append({"artist_name": artist_name,
                                       "recording_name": recording_name,
                                       "recording_mbid": data["recording_mbid"],
                                       "index": data["index"]})

                if artist_name != data["artist_name"]:
                    query_data.append({"artist_name": artist_name,
                                       "recording_name": recording_name,
                                       "recording_mbid": data["recording_mbid"],
                                       "index": data["index"]})

            # If nothing got cleaned, we can finish now
            if len(query_data) == 0:
                break

        ur = UnresolvedRecordingTracker()
        ur.add(unresolved_recording_mbids)

        return resolved_recordings

    def resolve_recording_by_mbid(self, artist_recording_data):
        """
            Given artist_recording_data, check to see if any of the recording MBIDs are
            in the local collection. If so, load the recording.id for it so it can be
            skipped later.
        """

        recording_index = {}
        for r in artist_recording_data:
            if "recording_mbid" in r:
                recording_index[r["recording_mbid"]] = r

        recording_mbids = list(recording_index.keys())
        recordings = Recording \
            .select(Recording) \
            .where(Recording.recording_mbid.in_(recording_mbids)) \
            .dicts()

        for recording in recordings:
            if recording["recording_mbid"] in recording_index:
                recording_index[recording["recording_mbid"]]["recording_id"] = recording["id"]

        return artist_recording_data

    def resolve_playlist(self, match_threshold, playlist):
        """
            Given a Troi playlist element, resolve tracks in the given playlist and update the playlist accordingly.
            threshold is a value between 0 and 1.0 for the percentage score required before a track is matched.
        """

        artist_recording_data = []
        # Check to make sure we have at least one recording
        try:
            _ = playlist.playlists[0].recordings[0]
        except (KeyError, IndexError):
            return playlist

        for rec in playlist.playlists[0].recordings:
            artist_recording_data.append({"artist_name": rec.artist_credit.name,
                                          "recording_name": rec.name,
                                          "recording_mbid": rec.mbid})

        # See what we can resolve using MBIDs
        artist_recording_data = self.resolve_recording_by_mbid(artist_recording_data)

        self.build_index()

        # Now see what we can resolve using fuzzy index
        hits = self.resolve_recordings(artist_recording_data, match_threshold)
        hit_index = {hit["index"]: hit for hit in hits}

        # load local recordings according to fuzzy search results
        recording_ids = [r["recording_id"] for r in hits]
        local_recordings = Recording \
            .select(Recording) \
            .where(Recording.id.in_(recording_ids)) \
            .dicts()

        # Build index based on recording.id
        rec_index = {r["id"]: r for r in local_recordings}

        logger.info("       %-40s %-40s %-40s" % ("RECORDING", "RELEASE", "ARTIST"))
        unresolved_recordings = []
        target_recordings = playlist.playlists[0].recordings
        resolved = 0
        failed = 0
        for i, artist_recording in enumerate(artist_recording_data):
            if i not in hit_index:
                logger.info(bcolors.FAIL + "FAIL " + bcolors.ENDC + "  %-40s %-40s %-40s" % (artist_recording["recording_name"][:39], "",
                                                                                           artist_recording["artist_name"][:39]))
                unresolved_recordings.append(artist_recording["recording_mbid"])
                failed += 1
                continue

            hit = hit_index[i]
            local_recording = rec_index[hit["recording_id"]]   # type content resolver recording
            target = target_recordings[hit["index"]]           # troi recordings
        
            if local_recording["file_id_type"] == FileIdType.FILE_PATH:
                target.musicbrainz["filename"] = local_recording["file_id"]
            if local_recording["file_id_type"] == FileIdType.SUBSONIC_ID:
                target.musicbrainz["subsonic_id"] = local_recording["file_id"]
            if local_recording["duration"] is not None:
                target.duration = local_recording["duration"]

            if not self.quiet:
                logger.info(bcolors.OKGREEN + ("%-5s" % hit["method"]) + bcolors.ENDC +
                            "  %-40s %-40s %-40s" % (artist_recording["recording_name"][:39], "",
                                                     artist_recording["artist_name"][:39]))
                logger.info("       %-40s %-40s %-40s" % (local_recording["recording_name"][:39],
                                                          local_recording["release_name"][:39],
                                                          local_recording["artist_name"][:39]))
            resolved += 1

        ur = UnresolvedRecordingTracker()
        ur.add(unresolved_recordings)

        if resolved == 0:
            logger.info("Sorry, but no tracks could be resolved, no playlist generated.")
            return []

        if not self.quiet:
            logger.info(f'\n{resolved} recordings resolved, {failed} not resolved.')

        return playlist



from peewee import *
from troi.content_resolver.model.database import db


class Directory(Model):
    """
    Basic metadata information about a recording on disk (a track).
    """

    class Meta:
        database = db

    id = AutoField()
    dir_path = TextField(null=False, unique=True)
    mtime = TimestampField(null=False)

    def __repr__(self):
        return "<Directory('%s')>" % self.dir_path



import unittest

import requests_mock

from troi.patch import Patch
from troi import Recording, Element, Playlist
from troi.tools.spotify_lookup import SPOTIFY_IDS_LOOKUP_URL


class DummyElement(Element):
    """ Dummy element that returns a fixed playlist for testing """

    @staticmethod
    def outputs():
        return [Playlist]

    def read(self, sources):
        recordings = [
            Recording(name="I Want It That Way", mbid="87dfa566-21c3-45ed-bc42-1d345b8563fa"),
            Recording(name="Untouchable"),
            Recording(name="The Perfect Kiss", mbid="ec0da94e-fbfe-4eb0-968e-024d4c32d1d0"),
            Recording(name="Love Your Voice", mbid="93726547-f8c0-4efd-8e16-d2dee76500f6"),
            Recording(name="Hall of Fame", mbid="395bd5a1-79cc-4e04-8869-ca9eabc78d09"),
        ]
        return [
            Playlist(
                name="Test Export Playlist",
                description="A playlist to test exporting playlists to spotify",
                recordings=recordings
            )
        ]


class DummyPatch(Patch):
    """ Dummy patch that always returns a fixed set of recordings for testing """

    @staticmethod
    def slug():
        return "test-patch"

    def create(self, inputs):
        return DummyElement()

    @staticmethod
    def outputs():
        return [Recording]


class TestSpotifySubmission(unittest.TestCase):

    @requests_mock.Mocker()
    def test_submit_to_spotify(self, mock_requests):
        self.maxDiff = None
        playlist_id = "33DUxaq2HQI7PDFODpFWJV"
        playlist_url = f"https://open.spotify.com/playlist/{playlist_id}"

        mock_requests.post(SPOTIFY_IDS_LOOKUP_URL, json=[
            {
                "artist_name": "Backstreet Boys",
                "recording_mbid": "87dfa566-21c3-45ed-bc42-1d345b8563fa",
                "release_name": "Millennium",
                "spotify_track_ids": ["47BBI51FKFwOMlIiX6m8ya"],
                "track_name": "I Want It That Way"
            },
            {
                "artist_name": "New Order",
                "recording_mbid": "ec0da94e-fbfe-4eb0-968e-024d4c32d1d0",
                "release_name": "Low‐Life",
                "spotify_track_ids": ["7y9bltr6hV3CsbqXWgwVZv", "4LWQfAhwP1Tf1wbzmT6NwW", "7clvpmRL6Ga8OyOs0is5RP"],
                "track_name": "The Perfect Kiss"
            },
            {
                "artist_name": "JONY",
                "recording_mbid": "93726547-f8c0-4efd-8e16-d2dee76500f6",
                "release_name": "Список твоих мыслей",
                "spotify_track_ids": ["4hyVrAsoKKjxAvQjPRt0ai", "75VNPADAFa4iNpjEBUYMhF"],
                "track_name": "Love Your Voice"
            },
            {
                "artist_name": "The Script featuring will.i.am",
                "recording_mbid": "395bd5a1-79cc-4e04-8869-ca9eabc78d09",
                "release_name": "#3",
                "spotify_track_ids": [],
                "track_name": "Hall of Fame"
            }
        ])

        mock_requests.post("https://api.spotify.com/v1/users/test-user-id/playlists", json={
            "id": playlist_id,
            "external_urls": {
                "spotify": playlist_url
            }
        })

        mock_requests.put(f"https://api.spotify.com/v1/playlists/{playlist_id}/tracks", json={"snapshot_id": "baz"})

        mock_requests.post(f"https://api.spotify.com/v1/playlists/{playlist_id}/tracks", json={"snapshot_id": "foo"})

        mock_requests.get(f"https://api.spotify.com/v1/playlists/{playlist_id}/tracks", json={
            "items": [
                {
                    "track": {
                        "name": "Backstreet Boys",
                        "id": "47BBI51FKFwOMlIiX6m8ya",
                        "is_playable": True
                    }
                },
                {
                    "track": {
                        "name": "New Order",
                        "id": "7y9bltr6hV3CsbqXWgwVZv",
                        "is_playable": False
                    }
                },
                {
                    "track": {
                        "name": "Love Your Voice",
                        "id": "4hyVrAsoKKjxAvQjPRt0ai",
                        "is_playable": True
                    }
                }
            ]
        })

        # the New Order track is unplayable and its alternative track ids should be rechecked
        mock_requests.get(
            f"https://api.spotify.com/v1/tracks/?ids=4LWQfAhwP1Tf1wbzmT6NwW,7clvpmRL6Ga8OyOs0is5RP&market=from_token",
            complete_qs=True,  # complete_qs to ensure that the track ids being rechecked are the exact ones we want
            json={
                "tracks": [
                    {
                        "id": "4LWQfAhwP1Tf1wbzmT6NwW",
                        "is_playable": True
                    },
                    {
                        "id": "7clvpmRL6Ga8OyOs0is5RP",
                        "is_playable": True
                    }
                ]
            }
        )
        mock_requests.put(f"https://api.spotify.com/v1/playlists/{playlist_id}/tracks", json={"snapshot_id": "bar"})

        patch = DummyPatch({
            "min_recordings": 1,
            "spotify": {
                "token": "spotify-auth-token",
                "user_id": "test-user-id",
                "is_public": True,
                "is_collaborative": False
            },
            "name": "Test Export Playlist",
            "upload": True
        })
        playlist = patch.generate_playlist()

        history = mock_requests.request_history

        # track 2 in playlist doesn't have a mbid and is skipped from request to lookup track ids
        self.assertEqual(history[0].json(), [
            {"recording_mbid": "87dfa566-21c3-45ed-bc42-1d345b8563fa"},
            {"recording_mbid": "ec0da94e-fbfe-4eb0-968e-024d4c32d1d0"},
            {"recording_mbid": "93726547-f8c0-4efd-8e16-d2dee76500f6"},
            {"recording_mbid": "395bd5a1-79cc-4e04-8869-ca9eabc78d09"}
        ])

        self.assertEqual(history[1].json(), {
            "name": "Test Export Playlist",
            "description": "A playlist to test exporting playlists to spotify",
            "public": True,
            "collaborative": False
        })

        # we had 4 tracks with mbids, out of those no matching spotify id was found for the last track id.
        # so we submitted the first spotify track id for the remaining tracks preserving the order in which
        # occur in the original playlist
        self.assertEqual(history[2].json(), [
            "spotify:track:47BBI51FKFwOMlIiX6m8ya",
            "spotify:track:7y9bltr6hV3CsbqXWgwVZv",
            "spotify:track:4hyVrAsoKKjxAvQjPRt0ai"
        ])

        # history[3] is the request to retrieve tracks for checking whether the playlist has unplayable tracks
        # history[4] is the request to retrieve info for alternative spotify ids of unplayable tracks, only thing
        # to check about this is the query strings which the request matcher in the beginning of this method does
        # history[5] is the request to reset the playlist tracks

        # we had found one track to be unplayable and replaced it with an alternative playable track. now updating the
        # playlist again so check the all the correct tracks and the new replaced tracks are sent and also in the
        # original order in which the tracks occur in the playlist
        self.assertEqual(history[6].json(), [
            "spotify:track:47BBI51FKFwOMlIiX6m8ya",
            "spotify:track:4LWQfAhwP1Tf1wbzmT6NwW",
            "spotify:track:4hyVrAsoKKjxAvQjPRt0ai"
        ])

        self.assertEqual(playlist.playlists[0].additional_metadata["external_urls"]["spotify"], playlist_url)



import logging

from troi.content_resolver.lb_radio import ListenBrainzRadioLocal
from troi.patches.periodic_jams_local import PeriodicJamsLocalPatch

logger = logging.getLogger(__name__)


class PeriodicJamsLocal(ListenBrainzRadioLocal):
    '''
       Generate local playlists against a music collection available via subsonic.
    '''

    def __init__(self, user_name, match_threshold, quiet):
        ListenBrainzRadioLocal.__init__(self, quiet)
        self.user_name = user_name
        self.match_threshold = match_threshold
        self.quiet = quiet

    def generate(self):
        """
           Generate a periodic jams playlist
        """
    
        patch = PeriodicJamsLocalPatch({
            "user_name": self.user_name,
            "quiet": self.quiet,
            "debug": True,
            "min_recordings": 1
        })

        # Now generate the playlist
        try:
            playlist = patch.generate_playlist()
        except RuntimeError as err:
            logger.info(f"LB Radio generation failed: {err}")
            return None

        if playlist is None:
            logger.info("Your prompt generated an empty playlist.")
            return {"playlist": {"track": []}}

        # Resolve any tracks that have not been resolved to a subsonic_id or a local file
        self.resolve_playlist(self.match_threshold, playlist)

        return playlist



import mutagen.oggvorbis

from troi.content_resolver.formats.tag_utils import get_tag_value, extract_track_number


EXTENSIONS = {'.ogg'}
READER = mutagen.oggvorbis.OggVorbis


def get_metadata(tags):

    mdata = {}
    mdata["artist_name"] = get_tag_value(tags, "artist")
    mdata["artist_sortname"] = get_tag_value(tags, "artistsort")
    mdata["release_name"] = get_tag_value(tags, "album")
    mdata["recording_name"] = get_tag_value(tags, "title")
    mdata["track_num"] = extract_track_number(get_tag_value(tags, "tracknumber"))
    mdata["disc_num"] = int(get_tag_value(tags, "discnumber") or 1)
    mdata["artist_mbid"] = get_tag_value(tags, "musicbrainz_artistid", "")
    mdata["recording_mbid"] = get_tag_value(tags, "musicbrainz_trackid", "")
    mdata["release_mbid"] = get_tag_value(tags, "musicbrainz_albumid", "")
    mdata["duration"] = int(tags.info.length * 1000)

    return mdata



import troi
from time import sleep
from random import shuffle

import requests

from troi import Recording
from troi import TARGET_NUMBER_OF_RECORDINGS


class LBRadioPlaylistRecordingElement(troi.Element):
    """
        Given an LB playlist, fetch its tracks and randomly include recordiungs from it. mode does not
        apply to this element.
    """

    NUM_RECORDINGS_TO_COLLECT = TARGET_NUMBER_OF_RECORDINGS * 2

    def __init__(self, mbid, mode="easy"):
        troi.Element.__init__(self)
        self.mbid = mbid
        self.mode = mode

    def inputs(self):
        return []

    def outputs(self):
        return [Recording]

    def read(self, entities):

        # Fetch the playlist
        while True:
            r = requests.get(f"https://api.listenbrainz.org/1/playlist/{self.mbid}")
            if r.status_code == 404:
                raise RuntimeError(f"Cannot find playlist {self.mbid}.")
            if r.status_code == 429:
                sleep(2)
                continue
            if r.status_code != 200:
                raise RuntimeError(f"Cannot fetch playlist {self.mbid}. {r.text}")

            break

        # Give feedback about the playlist
        self.local_storage["data_cache"]["element-descriptions"].append(f"playlist {self.mbid}")

        # Fetch the recordings, then shuffle
        mbid_list = []
        for recording in r.json()["playlist"]["track"]:
            identifiers = recording["identifier"]
            if isinstance(identifiers, str):
                identifiers = [identifiers]

            mbid = None
            for identifier in identifiers:
                if identifier.startswith("https://musicbrainz.org/recording/") or \
                        identifier.startswith("http://musicbrainz.org/recording/"):
                    mbid = identifier.split("/")[-1]
                    break

            mbid_list.append(mbid)

        shuffle(mbid_list)

        # Select and convert the first n MBIDs into Recordings
        recordings = []
        for mbid in mbid_list[:self.NUM_RECORDINGS_TO_COLLECT]:
            recordings.append(Recording(mbid=mbid))

        return recordings



import logging

import troi
from abc import ABC, abstractmethod

from troi.playlist import PlaylistElement
from troi.logging_utils import set_log_level
from troi.recording_search_service import RecordingSearchByTagService, RecordingSearchByArtistService

logger = logging.getLogger(__name__)

default_patch_args = dict(save=False,
                          token=None,
                          upload=False,
                          args=None,
                          created_for=None,
                          name=None,
                          desc=None,
                          min_recordings=10,
                          spotify=None,
                          apple_music=None,
                          soundcloud=None,
                          quiet=False)


class Patch(ABC):

    def __init__(self, args):
        self.quiet = False
        self.args = args

        # Dict used for local storage
        self.local_storage = {}

        self.patch_args = {**default_patch_args, **args}
        self.pipeline = self.create(self.patch_args)
        self._set_element_patch(self.pipeline)

        # Setup extensible services
        self.services = {}
        self.register_service(RecordingSearchByTagService())
        self.register_service(RecordingSearchByArtistService())

    @staticmethod
    def inputs():
        """
            This function should return a list of dicts that defined the type (argument or option), args, and kwargs to be passed to the click function. MusicBrainz entities and python base types can all be used. The documentation of the method is used as the help returned by the command. Example:

            .. code-block:: json

                [
                    {
                        "type" : "argument",
                        "args": ["num_recordings"],
                        "kwargs": {
                            "optional": true
                        }
                    }
                ]

        """
        return None

    @staticmethod
    def slug():
        ''' 
            Return the slug for this patch -- this is a URL friendly short identifier that can be used to invoke this patch via an HTTP call.

            e.g area-random-recordings
        '''

        return None

    @staticmethod
    def description():
        '''
            Return the description for this patch -- this short description (not more than a paragraph) should give the user an idea as to what the patch does.

            e.g "Generate a list of random recordings from a given area."
        '''

        return None

    @abstractmethod
    def create(self, input_args):
        """
            The function creates the data pipeline and then returns it. 

            Params:
               input_args: the arguments passed to the patch.
        """
        return None

    def register_service(self, service):
        """
            Register a new service that can provide services to troi patches.

            Only one service can be registered for any given service slug at a time. The most
            recently registered service will be use for the next playlist generation.
        """
        self.services[service.slug] = service

    def get_service(self, slug):
        """
           Given a service slug, return the class registered for this service. 

           Raises IndexError if no such service is registered.
        """
        return self.services[slug]

    def is_local(self):
        """
            If this function returns True, it means that the patch expects to use the local database, so Troi should
            setup the local database before running this patch. Returns False unless overridden by a deriving patch.
        """
        return False

    def post_process(self):
        """
            This function is called once the pipeline has produced its playlist, just before the Playlist object is created.
            This function could be used to inspect data in patch local storage to create the detailed playlist name
            and descriptionm which may not be available when the pipeline is constructed.
        """
        return

    def user_feedback(self):
        """
          Call this function to retrieve a list of strings that give the user feedback about the playlist that was generated, if any.
        """
        try:
            return self.local_storage["user_feedback"]
        except KeyError:
            return []

    def generate_playlist(self):
        """
        Generate a playlist 

        The args parameter is a dict and may containt the following keys:

        * quiet: Do not print out anything
        * save: The save option causes the generated playlist to be saved to disk.
        * token: Auth token to use when using the LB API. Required for submitting playlists to the server. See https://listenbrainz.org/profile to get your user token.
        * upload: Whether or not to submit the finished playlist to the LB server. Token must be set for this to work.
        * created-for: If this option is specified, it must give a valid user name and the TOKEN argument must specify a user who is whitelisted as a playlist bot at listenbrainz.org .
        * name: Override the algorithms that generate a playlist name and use this name instead.
        * desc: Override the algorithms that generate a playlist description and use this description instead.
        * min-recordings: The minimum number of recordings that must be present in a playlist to consider it complete. If it doesn't have sufficient numbers of tracks, ignore the playlist and don't submit it. Default: Off, a playlist with at least one track will be considere complete.
        * spotify: if present, attempt to submit the playlist to spotify as well. should be a dict and contain the spotify user id, spotify auth token with appropriate permissions, whether the playlist should be public, private or collaborative. it can also optionally have the existing urls to update playlists instead of creating new ones.
        * apple_music: if present, attempt to submit the playlist to Apple Music as well. should be a dict and contain the apple developer token, user music token, whether the playlist should be public, private. it can also optionally have the existing urls to update playlists instead of creating new ones.
        * soundcloud: if present, attempt to submit the palylist to soundcloud. should contain soundcloud auth token, whether the playlist should be public or private
        """

        try:
            set_log_level(self.patch_args.get("quiet", False))
            playlist = PlaylistElement()
            playlist.set_sources(self.pipeline)
            logger.info("Troi playlist generation starting...")
            result = playlist.generate(self.quiet)

            name = self.patch_args["name"]
            if name:
                playlist.playlists[0].name = name

            desc = self.patch_args["desc"]
            if desc:
                playlist.playlists[0].descripton = desc

            logger.info("done.")
        except troi.PipelineError as err:
            raise RuntimeError("Playlist generation failed: %s" % err)

        upload = self.patch_args["upload"]
        token = self.patch_args["token"]
        spotify = self.patch_args["spotify"]
        apple_music = self.patch_args["apple_music"]
        soundcloud = self.patch_args["soundcloud"]
        if upload and not token and not spotify and not apple_music and not soundcloud:
            raise RuntimeError("In order to upload a playlist, you must provide an auth token. Use option --token.")

        min_recordings = self.patch_args["min_recordings"]
        if min_recordings is not None and \
                (len(playlist.playlists) == 0 or len(playlist.playlists[0].recordings) < min_recordings):
            raise RuntimeError("Playlist does not have at least %d recordings" % min_recordings)

        save = self.patch_args["save"]
        if result is not None and spotify and upload:
            for url, _ in playlist.submit_to_spotify(spotify["user_id"], spotify["token"], spotify["is_public"],
                                                     spotify["is_collaborative"], spotify.get("existing_urls", [])):
                logger.info("Submitted playlist to spotify: %s" % url)

        if result is not None and soundcloud and upload:
            for url, _ in playlist.submit_to_soundcloud(soundcloud["user_id"], soundcloud["token"], soundcloud["is_public"],
                                                soundcloud.get("existing_urls", [])):
                logger.info("Submitted playlist to soundcloud: %s" % url)

        if result is not None and apple_music and upload:
            for url, _ in playlist.submit_to_apple_music(apple_music["music_user_token"], apple_music["developer_token"], apple_music["is_public"], apple_music.get("existing_urls", [])):
                logger.info("Submitted playlist to apple music: %s" % url)

        created_for = self.patch_args["created_for"]
        if result is not None and token and upload:
            for url, _ in playlist.submit(token, created_for):
                logger.info("Submitted playlist: %s" % url)

        if result is not None and save:
            playlist.save()
            logger.info("playlist saved.")

        if not self.quiet and result is not None:
            logger.info("")
            playlist.print()

        if len(playlist.playlists) == 0:
            logger.info("No playlists were generated. :(")
        elif len(playlist.playlists) == 1:
            logger.info("A playlist with %d tracks was generated." % len(playlist.playlists[0].recordings))
        else:
            logger.info("%d playlists were generated." % len(playlist.playlists))

        return playlist

    def _set_element_patch(self, element):
        """ 
            Go through the pipeline, setting the patch objects for each Element
        """
        element.set_patch_object(self)
        for src in element.sources:
            self._set_element_patch(src)



import datetime
from peewee import *
from troi.content_resolver.model.database import db
from troi.content_resolver.model.recording import Recording


class Tag(Model):
    """
       Represents a tag that could be joined to an entity
    """

    class Meta:
        database = db

    id = AutoField()
    name = TextField(null=False, unique=True)

    def __repr__(self):
        return "<Tag('%s')>" % (self.name or "")


class RecordingTag(Model):
    """
      A tag connected to a recording
    """

    class Meta:
        database = db
        table_name = "recording_tag"

    id = AutoField()
    recording = ForeignKeyField(Recording)
    tag = ForeignKeyField(Tag, on_delete="CASCADE")
    last_updated = DateTimeField(null=False, default=datetime.datetime.now)
    entity = TextField(null=False)

    def __repr__(self):
        return "<RecordingTag('%s','%d')>" % (self.tag.name or "", self.recording)



import troi
from troi import Recording
from troi.musicbrainz.mbid_reader import MBIDReaderElement
from troi.musicbrainz.recording_lookup import RecordingLookupElement
from troi.patch import Patch
from troi.playlist import PlaylistMakerElement


class PlaylistFromMBIDsPatch(Patch):
    """
    """

    def __init__(self, args):
        super().__init__(args)

    @staticmethod
    def inputs():
        """
        Make a playlist from a file containing one MBID per line.

        \b
        FILE_NAME: filename that contains MBIDS
        """
        return [{"type": "argument", "args": ["file_name"]}]

    @staticmethod
    def outputs():
        return [Recording]

    @staticmethod
    def slug():
        return "playlist-from-mbids"

    @staticmethod
    def description():
        return "Generate a playlist from a list of MBIDSs"

    def create(self, inputs):

        source = MBIDReaderElement(inputs['file_name'])

        rec_lookup = RecordingLookupElement()
        rec_lookup.set_sources(source)

        pl_maker = PlaylistMakerElement("Playlist made from MBIDs", "", patch_slug=self.slug())
        pl_maker.set_sources(rec_lookup)

        return pl_maker



#!/usr/bin/env python3
import logging
import sys

import click

import troi
import troi.playlist
import troi.utils


logger = logging.getLogger(__name__)


def list_patches():
    """
        Print a list of all available patches
    """
    patches = troi.utils.discover_patches()

    logger.info("Available patches:")
    size = max([len(k) for k in patches])
    for slug in sorted(patches or []):
        patch = patches[slug]
        print("%s:%s %s" % (slug, " " * (size - len(slug)), patch.description()))


def patch_info(patch):
    """
        Get info for a given patch

        :param patch: the patch to get info for.
    """

    patches = troi.utils.discover_patches()
    if patch not in patches:
        logger.error("Cannot load patch '%s'. Use the list command to get a list of available patches." % patch)
        sys.exit(1)

    apatch = patches[patch]
    cmd = convert_patch_to_command(apatch)
    context = click.Context(cmd, info_name=patch)
    click.echo(cmd.get_help(context))


def convert_patch_to_command(patch):
    """
        Convert patch object to dummy click command to parse args and show help

        :param patch: the patch to get info for.
    """

    def f(**data):
        return data

    f.__doc__ = patch.inputs.__doc__

    for _input in reversed(patch.inputs()):
        args = _input.get("args", [])
        kwargs = _input.get("kwargs", {})
        if _input["type"] == "argument":
            f = click.argument(*args, **kwargs)(f)
        elif _input["type"] == "option":
            f = click.option(*args, **kwargs)(f)
        else:
            click.echo("Patch is invalid, contact patch writer to fix")

    return click.command(no_args_is_help=True)(f)



from datetime import datetime, timedelta

import troi.listenbrainz.recs
import troi.musicbrainz.recording_lookup
from troi import Playlist
from troi.patch import Patch
from troi.playlist import PlaylistMakerElement

from troi.local.recording_resolver import RecordingResolverElement

DAYS_OF_RECENT_LISTENS_TO_EXCLUDE = 60  # Exclude tracks listened in last X days from the daily jams playlist
DAILY_JAMS_MIN_RECORDINGS = 25  # the minimum number of recordings we aspire to have in a daily jam, this is not a hard limit
BATCH_SIZE_RECS = 1000  # the number of recommendations fetched in 1 go
MAX_RECS_LIMIT = 1000  # the maximum of recommendations available in LB


class PeriodicJamsLocalPatch(Patch):
    """
    """

    def __init__(self, args):
        super().__init__(args)

    @staticmethod
    def inputs():
        """
        Generate a periodic playlist from the ListenBrainz recommended recordings.

        \b
        USER_NAME is a MusicBrainz user name that has an account on ListenBrainz.
        """
        return [{
            "type": "argument",
            "args": ["user_name"]
        }]

    @staticmethod
    def outputs():
        return [Playlist]

    @staticmethod
    def slug():
        return "periodic-jams-local"

    @staticmethod
    def description():
        return "Generate a localized periodic playlist from the ListenBrainz recommended recordings."

    def is_local(self):
        return True

    def create(self, inputs):
        user_name = inputs['user_name']

        recs = troi.listenbrainz.recs.UserRecordingRecommendationsElement(user_name,
                                                                          "raw",
                                                                          count=1000)

        recent_listens_lookup = troi.listenbrainz.listens.RecentListensTimestampLookup(user_name,
                                                                                       days=2)
        recent_listens_lookup.set_sources(recs)

        latest_filter = troi.filters.LatestListenedAtFilterElement(DAYS_OF_RECENT_LISTENS_TO_EXCLUDE)
        latest_filter.set_sources(recent_listens_lookup)

        feedback_lookup = troi.listenbrainz.feedback.ListensFeedbackLookup(user_name, auth_token=inputs.get("token"))
        feedback_lookup.set_sources(latest_filter)

        recs_lookup = troi.musicbrainz.recording_lookup.RecordingLookupElement()
        recs_lookup.set_sources(feedback_lookup)

        resolve = RecordingResolverElement(.8, self.quiet)
        resolve.set_sources(recs_lookup)

        pl_maker = PlaylistMakerElement(name="Weekly Jams for %s" % (user_name),
                                        desc="test playlist!",
                                        patch_slug="periodic-jams",
                                        max_num_recordings=50,
                                        max_artist_occurrence=2,
                                        shuffle=True,
                                        expires_at=datetime.utcnow() + timedelta(weeks=2))
        pl_maker.set_sources(resolve)

        return pl_maker



import troi
from random import randint, shuffle
from uuid import UUID

import troi
import liblistenbrainz
import liblistenbrainz.errors
from troi import Artist, Recording
from troi import TARGET_NUMBER_OF_RECORDINGS
from troi.parse_prompt import TIME_RANGES


class LBRadioRecommendationRecordingElement(troi.Element):
    """
        Given a LB user, fetch their recommended recordings and then include recordings from it.
    """

    MAX_RECOMMENDED_RECORDINGS = 1000
    MAX_RECORDINGS_TO_FETCH_PER_CALL = 100

    def __init__(self, user_name, listened="all", mode="easy"):
        troi.Element.__init__(self)
        self.user_name = user_name
        self.listened = listened
        self.mode = mode
        self.client = liblistenbrainz.ListenBrainz()

    def inputs(self):
        return []

    def outputs(self):
        return [Recording]

    def read(self, entities):

        if self.mode == "easy":
            offset = 0
        elif self.mode == "medium":
            offset = self.MAX_RECOMMENDED_RECORDINGS // 3
        else:
            offset = self.MAX_RECOMMENDED_RECORDINGS * 2 // 3

        added = 0
        skipped = 0
        recordings = []
        count = self.MAX_RECOMMENDED_RECORDINGS // 3
        while count > 0:
            # Fetch the user recs
            try:
                result = self.client.get_user_recommendation_recordings(self.user_name, "raw",
                                                                        min(self.MAX_RECORDINGS_TO_FETCH_PER_CALL, count), offset)
            except liblistenbrainz.errors.ListenBrainzAPIException as err:
                raise RuntimeError("Cannot fetch recording stats for user %s" % self.user_name)

            if result is None or len(result['payload']['mbids']) == 0:
                break

            # Turn them into recordings
            for r in result['payload']['mbids']:
                if r['recording_mbid'] is not None:
                    offset += 1
                    latest = r.get("latest_listened_at", None)
                    if self.listened == "all" or (self.listened == "unlistened" and latest is None) or \
                            (self.listened == "listened" and latest is not None):
                        count -= 1
                        recordings.append(Recording(mbid=r['recording_mbid']))
                        added += 1
                    else:
                        skipped += 1

            # Shuffle the recordings
            shuffle(recordings)

        # Give feedback on what we collected
        listened = ""
        if self.listened != "all":
            listened = f"previously {self.listened} "

        self.local_storage["data_cache"]["element-descriptions"].append(f"{self.user_name}'s {listened}recommended songs")

        # TODO: How do we prevent sequential tracks by the same artist?

        return recordings



from uuid import UUID
import unittest

from troi.parse_prompt import PromptParser, ParseError


class TestParser(unittest.TestCase):

    def test_basic_entities(self):
        pp = PromptParser()
        r = pp.parse("artist:(57baa3c6-ee43-4db3-9e6a-50bbc9792ee4)")
        assert r[0] == {"entity": "artist", "values": [UUID("57baa3c6-ee43-4db3-9e6a-50bbc9792ee4")], "weight": 1, "opts": []}

        r = pp.parse("artist:(57baa3c6-ee43-4db3-9e6a-50bbc9792ee4)")
        assert r[0] == {"entity": "artist", "values": [UUID("57baa3c6-ee43-4db3-9e6a-50bbc9792ee4")], "weight": 1, "opts": []}

        self.assertRaises(ParseError, pp.parse, "wrong:57baa3c6-ee43-4db3-9e6a-50bbc9792ee4")
        self.assertRaises(ParseError, pp.parse, "artist:57baa3c6-ee43-4db3-9e6a-50bbc9792ee4")

        r = pp.parse("artist:(the knife)")
        assert r[0] == {"entity": "artist", "values": ["the knife"], "weight": 1, "opts": []}

        self.assertRaises(ParseError, pp.parse, "artist:u2:nosim")
        self.assertRaises(ParseError, pp.parse, "artists:u2:nosim")
        self.assertRaises(ParseError, pp.parse, "country:andorra")

    def test_tags(self):
        pp = PromptParser()
        r = pp.parse("tag:abstract tag:rock tag:blues")
        assert r[0] == {"entity": "tag", "values": ["abstract"], "weight": 1, "opts": []}
        assert r[1] == {"entity": "tag", "values": ["rock"], "weight": 1, "opts": []}
        assert r[2] == {"entity": "tag", "values": ["blues"], "weight": 1, "opts": []}

        r = pp.parse("tag:(abstract,rock,blues)")
        assert r[0] == {"entity": "tag", "values": ["abstract", "rock", "blues"], "weight": 1, "opts": []}

        r = pp.parse("tag:(abstract rock blues)")
        assert r[0] == {"entity": "tag", "values": ["abstract rock blues"], "weight": 1, "opts": []}

        r = pp.parse("tag:(abstract,rock,blues):1:or")
        assert r[0] == {"entity": "tag", "values": ["abstract", "rock", "blues"], "weight": 1, "opts": ["or"]}
        r = pp.parse("tag:(abstract,rock,blues)")
        assert r[0] == {"entity": "tag", "values": ["abstract", "rock", "blues"], "weight": 1, "opts": []}

        r = pp.parse('tag:(trip hop, hip hop)')
        assert r[0] == {"entity": "tag", "values": ["trip hop", "hip hop"], "weight": 1, "opts": []}

        r = pp.parse("tag:(r&b)")
        assert r[0] == {"entity": "tag", "values": ["r&b"], "weight": 1, "opts": []}

        r = pp.parse("tag:(blümchen)")
        assert r[0] == {"entity": "tag", "values": ["blümchen"], "weight": 1, "opts": []}

        r = pp.parse("tag:(モーニング娘。)")
        assert r[0] == {"entity": "tag", "values": ["モーニング娘。"], "weight": 1, "opts": []}

        r = pp.parse("tag:(57baa3c6-ee43-4db3-9e6a-50bbc9792ee4)")
        assert r[0] == {"entity": "tag", "values": ["57baa3c6-ee43-4db3-9e6a-50bbc9792ee4"], "weight": 1, "opts": []}

    def test_tag_errors(self):
        pp = PromptParser()
        self.assertRaises(ParseError, pp.parse, "t:(abstract rock blues):bork")
        self.assertRaises(ParseError, pp.parse, "tag:(foo")
        self.assertRaises(ParseError, pp.parse, "tag:foo)")

    def test_shortcuts(self):
        pp = PromptParser()
        r = pp.parse("#abstract")
        assert r[0] == {"entity": "tag", "values": ["abstract"], "weight": 1, "opts": []}

        pp = PromptParser()
        r = pp.parse("u2")
        assert r[0] == {"entity": "artist", "values": ["u2"], "weight": 1, "opts": []}

        pp = PromptParser()
        r = pp.parse("amy winehouse")
        assert r[0] == {"entity": "artist", "values": ["amy winehouse"], "weight": 1, "opts": []}

    def test_compound(self):
        pp = PromptParser()
        r = pp.parse('artist:(05319f96-e409-4199-b94f-3cabe7cc188a):2 tag:(downtempo):1 tag:(trip hop, abstract):2')
        assert r[0] == {"entity": "artist", "values": [UUID("05319f96-e409-4199-b94f-3cabe7cc188a")], "weight": 2, "opts": []}
        assert r[1] == {"entity": "tag", "values": ["downtempo"], "weight": 1, "opts": []}
        assert r[2] == {"entity": "tag", "values": ["trip hop", "abstract"], "weight": 2, "opts": []}

    def test_weights(self):
        pp = PromptParser()
        r = pp.parse("artist:(57baa3c6-ee43-4db3-9e6a-50bbc9792ee4):1 artist:(f54ba4c6-12dd-4358-9136-c64ad89420c5):2")
        assert r[0] == {"entity": "artist", "values": [UUID("57baa3c6-ee43-4db3-9e6a-50bbc9792ee4")], "weight": 1, "opts": []}
        assert r[1] == {"entity": "artist", "values": [UUID("f54ba4c6-12dd-4358-9136-c64ad89420c5")], "weight": 2, "opts": []}

        self.assertRaises(ParseError, pp.parse,
                          "artist:57baa3c6-ee43-4db3-9e6a-50bbc9792ee4:1 artist:f54ba4c6-12dd-4358-9136-c64ad89420c5:fussy")
        self.assertRaises(ParseError, pp.parse, "artist:57baa3c6-ee43-4db3-9e6a-50bbc9792ee4:1 artist:f54ba4c6-12dd-4358-9136-c64ad89420c5:.5")

        r = pp.parse("artist:(portishead)::easy")
        assert r[0] == {"entity": "artist", "values": ["portishead"], "weight": 1, "opts": ["easy"]}

        r = pp.parse("artist:(57baa3c6-ee43-4db3-9e6a-50bbc9792ee4)::easy")
        assert r[0] == {"entity": "artist", "values": [UUID("57baa3c6-ee43-4db3-9e6a-50bbc9792ee4")], "weight": 1, "opts": ["easy"]}

    def test_opts(self):
        pp = PromptParser()
        r = pp.parse("stats:(mr_monkey)::month")
        print(r[0])
        assert r[0] == {"entity": "stats", "values": ["mr_monkey"], "weight": 1, "opts": ["month"]}
        r = pp.parse("artist:(monkey)::nosim,easy")
        assert r[0] == {"entity": "artist", "values": ["monkey"], "weight": 1, "opts": ["nosim", "easy"]}

        self.assertRaises(ParseError, pp.parse, 'artist:(meh)::nosim,')

    def test_parens(self):
        pp = PromptParser()
        self.assertRaises(ParseError, pp.parse, 'artist:adfadf(meh)')
        self.assertRaises(ParseError, pp.parse, 'artist:adfadf(meh')
        self.assertRaises(ParseError, pp.parse, 'artist:adfadf)meh')

    def test_collection_playlist(self):
        pp = PromptParser()
        r = pp.parse("collection:(57baa3c6-ee43-4db3-9e6a-50bbc9792ee4)")
        assert r[0] == {"entity": "collection", "values": [UUID("57baa3c6-ee43-4db3-9e6a-50bbc9792ee4")], "weight": 1, "opts": []}

        r = pp.parse("playlist:(57baa3c6-ee43-4db3-9e6a-50bbc9792ee4)")
        assert r[0] == {"entity": "playlist", "values": [UUID("57baa3c6-ee43-4db3-9e6a-50bbc9792ee4")], "weight": 1, "opts": []}

        r = pp.parse("playlist:(57baa3c6-ee43-4db3-9e6a-50bbc9792ee4)")
        assert r[0] == {"entity": "playlist", "values": [UUID("57baa3c6-ee43-4db3-9e6a-50bbc9792ee4")], "weight": 1, "opts": []}

    def test_stats(self):
        pp = PromptParser()
        r = pp.parse("stats:mr_monkey:1:year")
        assert r[0] == {"entity": "stats", "values": ["mr_monkey"], "weight": 1, "opts": ["year"]}

        r = pp.parse("stats:rob:1:week")
        assert r[0] == {"entity": "stats", "values": ["rob"], "weight": 1, "opts": ["week"]}

        r = pp.parse("stats:(mr_monkey)::month")
        assert r[0] == {"entity": "stats", "values": ["mr_monkey"], "weight": 1, "opts": ["month"]}

        r = pp.parse("stats:(mr_monkey):2:month")
        assert r[0] == {"entity": "stats", "values": ["mr_monkey"], "weight": 2, "opts": ["month"]}

        r = pp.parse("stats:(rob zombie)")
        assert r[0] == {"entity": "stats", "values": ["rob zombie"], "weight": 1, "opts": []}

    def test_recs(self):
        pp = PromptParser()
        self.assertRaises(ParseError, pp.parse, 'recs:')

        r = pp.parse("recs:mr_monkey::listened")
        assert r[0] == {"entity": "recs", "values": ["mr_monkey"], "weight": 1, "opts": ["listened"]}

        r = pp.parse("recs:rob:1:unlistened")
        assert r[0] == {"entity": "recs", "values": ["rob"], "weight": 1, "opts": ["unlistened"]}

        r = pp.parse("recs:(mr_monkey):1:listened")
        assert r[0] == {"entity": "recs", "values": ["mr_monkey"], "weight": 1, "opts": ["listened"]}

        r = pp.parse("recs:(mr_monkey):2:unlistened")
        assert r[0] == {"entity": "recs", "values": ["mr_monkey"], "weight": 2, "opts": ["unlistened"]}

        r = pp.parse("recs:(rob zombie)")
        assert r[0] == {"entity": "recs", "values": ["rob zombie"], "weight": 1, "opts": []}

    def test_country(self):
        pp = PromptParser()
        r = pp.parse("country:(57baa3c6-ee43-4db3-9e6a-50bbc9792ee4)")
        assert r[0] == {"entity": "country", "values": [UUID("57baa3c6-ee43-4db3-9e6a-50bbc9792ee4")], "weight": 1, "opts": []}

        r = pp.parse("country:(mali)")
        assert r[0] == {"entity": "country", "values": ["mali"], "weight": 1, "opts": []}



from uuid import UUID
import re

TIME_RANGES = ["week", "month", "quarter", "half_yearly", "year", "all_time", "this_week", "this_month", "this_year"]
ELEMENTS = ["artist", "tag", "collection", "playlist", "stats", "recs", "country"]

ELEMENT_OPTIONS = {
    "artist": ["nosim", "easy", "medium", "hard"],
    "tag": ["nosim", "and", "or", "easy", "medium", "hard"],
    "collection": ["easy", "medium", "hard"],
    "playlist": ["easy", "medium", "hard"],
    "stats": TIME_RANGES,
    "recs": ["easy", "medium", "hard", "listened", "unlistened"],
    "country": ["easy", "medium", "hard"]
}

OPTIONS = set()
for eo in ELEMENT_OPTIONS:
    OPTIONS.update(ELEMENT_OPTIONS[eo])

OPTIONS = list(OPTIONS)


class ParseError(Exception):
    pass


class PromptParser:
    """
    Parse the LB radio prompt and return a list of elements and all of their data/options
    """

    def __init__(self):
        self.clean_spaces = re.compile(r"\s+")
        self.element_check = re.compile(r"([a-zA-Z]+):")

    def identify_block(self, block):
        """Given a prompt string, identify the block that is at the beginning of the string"""

        for element in ELEMENTS:
            if block.startswith(element + ":"):
                return element

        if block.startswith("#"):
            return "hashtag"

        # check for malformed element names
        m = self.element_check.match(block)
        if m is not None:
            raise ParseError("Unrecognized element name '%s'. Must be one of: %s" % (m.group(0), ",".join(ELEMENTS)))

        return "artistname"

    def parse_special_cases(self, prompt):
        """Detect the artist and tag special cases and re-write the query in long hand form."""

        block_type = self.identify_block(prompt)
        if block_type == "hashtag":
            return "tag:(%s)" % prompt[1:]
        if block_type == "artistname":
            return "artist:(%s)" % prompt

        return prompt

    def set_block_values(self, name, values, weight, opts, text, block):
        """Parse, process and sanity check data for an element"""

        if values is None:
            if name in ("artist", "country", "collection", "playlist"):
                try:
                    values = [UUID(text)]
                except ValueError:
                    values = [text]
            elif name == "tag":
                values = text.split(",")
                values = [v.strip() for v in values]
            else:
                values = [text]
        elif weight is None:
            if not text:
                weight = 1
            else:
                try:
                    weight = int(text)
                except ValueError:
                    raise ParseError("Weight must be a positive integer, not '%s'" % text)
        elif not opts:
            opts = text.split(",")
            if opts and opts[-1] == "":
                raise ParseError("Trailing comma in options.")

        return values, weight, opts

    def parse(self, prompt):
        """Parse an actual LB-radio prompt and return a list of [name, [values], weight, [opts]]"""

        prompt = self.clean_spaces.sub(" ", prompt).strip()
        block = self.parse_special_cases(prompt)
        blocks = []
        while True:
            name = ""
            for element in ELEMENTS:
                if block.startswith(element + ":"):
                    name = element
                    break
            if not name:
                raise ParseError("Unknown element '%s'" % block)

            block = block[len(name):]
            if block[0] == ':':
                block = block[1:]

            opts = []
            weight = None
            values = None
            text = ""
            parens = 0
            escaped = False
            for i in range(len(block)):
                if not escaped and block[i] == '\\':
                    escaped = True
                    continue

                if escaped:
                    if block[i] in ('(', ')', '\\'):
                        escaped = False
                        text += block[i]
                        continue
                    escaped = False

                if block[i] == '(':
                    if i > 0:
                        raise ParseError("() must start at the beginning of the value field.")
                    parens += 1
                    continue

                if block[i] == ')':
                    parens -= 1
                    if parens < 0:
                        raise ParseError("closing ) without matching opening ( near: '%s'." % block[i:])
                    continue

                if block[i] == ':' and parens == 0:
                    values, weight, opts = self.set_block_values(name, values, weight, opts, text, block)
                    text = ""
                    continue

                # Check to make sure that some values are in ()
                if name in ("artist", "country", "collection", "playlist") and i == 0 and not block[i] == "(":
                    raise ParseError("Element value must be enclosed in ( ). Try: %s:(name)" % (name))

                if block[i] == ' ' and parens == 0:
                    break

                text += block[i]

            # Now that we've parsed a block, do some sanity checking
            values, weight, opts = self.set_block_values(name, values, weight, opts, text, block)
            try:
                block = block[i + 1:]
            except UnboundLocalError:
                raise ParseError("incomplete prompt")

            if parens > 0:
                raise ParseError("Missing closing ).")

            if parens < 0:
                raise ParseError("Missing opening (.")

            for opt in opts:
                if opt not in ELEMENT_OPTIONS[name]:
                    raise ParseError("Option '%s' is not allowed for element %s" % (opt, name))

            blocks.append({"entity": name, "values": values, "weight": weight or 1, "opts": opts})

            if len(block) == 0:
                break

        return blocks



import logging
from time import sleep

import requests

import troi.patch
from troi import TARGET_NUMBER_OF_RECORDINGS
from troi.plist import plist
from troi import Element, ArtistCredit, Recording, PipelineError, DEVELOPMENT_SERVER_URL

logger = logging.getLogger(__name__)


class LBRadioCountryRecordingElement(Element):
    '''
        Given a country, return recordings for that country.

        Arguments:
            area_name: the name of the area to make a playlist for
    '''

    def __init__(self, mode, area_name=None, area_mbid=None):
        super().__init__()
        self.area_name = area_name
        self.area_mbid = str(area_mbid)
        self.mode = mode

    @staticmethod
    def inputs():
        return []

    @staticmethod
    def outputs():
        return [Recording]

    def lookup_area_by_name(self, area_name):

        while True:
            r = requests.get("https://musicbrainz.org/ws/2/area?query=%s&fmt=json" % area_name)
            if r.status_code in (503, 429):
                sleep(1)
                continue

            if r.status_code != 200:
                raise PipelineError("Cannot fetch country code from MusicBrainz. HTTP code %s" % r.status_code)

            try:
                area = r.json()['areas'][0]
            except IndexError:
                return None

            if area["type"] == "Country":
                return area["id"]
            else:
                return None

    def lookup_area_by_mbid(self, area_mbid):

        while True:
            r = requests.get("https://musicbrainz.org/ws/2/area/%s?fmt=json" % area_mbid)
            if r.status_code in (503, 429):
                sleep(1)
                continue

            if r.status_code != 200:
                raise PipelineError("Cannot fetch country code from MusicBrainz. Error: %s" % r.text)

            area = r.json()
            if area["type"] != "Country":
                raise PipelineError("The specified area_mbid (%s) refers to a %s, but only countries are supported." %
                                    (area_mbid, area["type"]))

            return area["name"]

    def recording_from_row(self, row):
        if row['recording_mbid'] is None:
            return None

        r = Recording(mbid=row['recording_mbid'])
        if 'artist_credit_name' in row:
            r.artist = ArtistCredit(name=row['artist_credit_name'])

        if 'recording_name' in row:
            r.name = row['recording_name']

        if 'year' in row:
            r.year = row['year']

        if 'listen_count' in row:
            r.listenbrainz = {"listen_count": row["listen_count"]}

        return r

    def read(self, inputs):

        start, stop = {"easy": (66, 100), "medium": (33, 66), "hard": (0, 33)}[self.mode]

        if self.area_name is None and self.area_mbid is None:
            raise PipelineError("An area name or area mbid must be specified.")

        if self.area_name:
            self.area_mbid = self.lookup_area_by_name(self.area_name)
            if self.area_mbid is None:
                raise PipelineError("Cannot find country '%s'" % self.area_name)
        else:
            self.area_name = self.lookup_area_by_mbid(self.area_mbid)
            if self.area_name is None:
                raise PipelineError("Cannot lookup country from mbid '%s'" % self.area_mbid)

        args = [{"[area_mbid]": self.area_mbid}]
        r = requests.post(DEVELOPMENT_SERVER_URL + "/popular-recordings-by-country/json", json=args)
        if r.status_code != 200:
            raise PipelineError("Cannot fetch first dataset recordings from ListenBrainz. HTTP code %s (%s)" %
                                (r.status_code, r.text))

        self.data_cache = self.local_storage["data_cache"]
        self.data_cache["element-descriptions"].append("country %s" % self.area_name)

        recordings = plist()
        for row in r.json():
            recordings.append(self.recording_from_row(row))

        return recordings.random_item(start, stop, TARGET_NUMBER_OF_RECORDINGS)



import unittest
from unittest.mock import patch

from troi import Artist, ArtistCredit, Release, Recording
import troi.listenbrainz.stats

artist_ret = {
    'payload': {
        'artists': [
            {
                 'artist_mbids': [], 
                 'artist_name': 'Pretty Lights',
                 'listen_count': 1641
            }
         ]
     }
}

release_ret = {
    'payload': {
        'count': 25, 
        'from_ts': 1009843200, 
        'last_updated': 1593870248, 
        'offset': 0, 
        'range': 'all_time', 
        'releases': [
            {
                'artist_mbids': [], 
                'artist_name': 'Saltillo',
                'listen_count': 633, 
                'release_mbid': None, 
                'release_name': 'Ganglion'
            }
        ]
    }
}

recording_ret = {
    'payload': {
        'count': 25, 
        'from_ts': 1009843200, 
        'last_updated': 1593872937, 
        'offset': 0, 
        'range': 'all_time', 
        'recordings': [
            {
                'artist_mbids': [], 
                'artist_name': 'Woolfy',
                'listen_count': 94, 
                'recording_mbid': None, 
                'release_mbid': None, 
                'release_name': 'Stations',
                'track_name': 'Tangiers'
            }
        ]
    }
}

class TestStats(unittest.TestCase):

    @patch('liblistenbrainz.ListenBrainz.get_user_artists')
    def test_user_artist_stats(self, stats_mock):

        stats_mock.return_value = artist_ret
        u = troi.listenbrainz.stats.UserArtistsElement("rob", 1, 1, "week")
        entities = u.read()
        stats_mock.assert_called_with("rob", 1, 1, "week")
        assert len(entities) == 1
        assert isinstance(entities[0], ArtistCredit)
        assert entities[0].name == 'Pretty Lights'
        assert len(entities[0].artists) == 0

    @patch('liblistenbrainz.ListenBrainz.get_user_releases')
    def test_user_release_stats(self, stats_mock):

        stats_mock.return_value = release_ret
        u = troi.listenbrainz.stats.UserReleasesElement("rob", 1, 1, "all_time")
        entities = u.read()
        stats_mock.assert_called_with("rob", 1, 1, "all_time")
        assert len(entities) == 1
        assert isinstance(entities[0], Release)
        assert entities[0].artist_credit.name == 'Saltillo'
        assert entities[0].name == 'Ganglion'

    @patch('liblistenbrainz.ListenBrainz.get_user_recordings')
    def test_user_recording_stats(self, stats_mock):

        stats_mock.return_value = recording_ret
        u = troi.listenbrainz.stats.UserRecordingElement("rob", 1, 1, "all_time")
        entities = u.read()
        stats_mock.assert_called_with("rob", 1, 1, "all_time")
        assert len(entities) == 1
        assert isinstance(entities[0], Recording)
        assert entities[0].artist_credit.name == 'Woolfy'
        assert entities[0].release.name == 'Stations'
        assert entities[0].name == 'Tangiers'



import logging
from collections import defaultdict, namedtuple
import datetime
from time import sleep

import requests
from tqdm import tqdm

from troi.content_resolver.model.database import db
from troi.content_resolver.model.recording import Recording, RecordingMetadata
from troi.content_resolver.model.tag import RecordingTag

logger = logging.getLogger(__name__)


RecordingRow = namedtuple('RecordingRow', ('id', 'mbid', 'metadata_id'))


class MetadataLookup:
    '''
    Given the local database, lookup metadata from MusicBrainz to allow local playlist resolution.
    '''

    BATCH_SIZE = 1000

    def __init__(self, quiet):
        self.quiet = quiet

    def lookup(self):
        """
        Iterate over all recordings in the database and call lookup_chunk for chunks of recordings.
        """

        cursor = db.execute_sql("""SELECT recording.id, recording.recording_mbid, recording_metadata.id
                                     FROM recording
                                LEFT JOIN recording_metadata
                                       ON recording.id = recording_metadata.recording_id
                                    WHERE recording_mbid IS NOT NULL
                                 ORDER BY artist_name, release_name""")
        recordings = tuple(
            RecordingRow(id=row[0], mbid=str(row[1]), metadata_id=row[2])
            for row in cursor.fetchall()
        )

        logger.info("[ %d recordings to lookup ]" % len(recordings))

        offset = 0

        if not self.quiet:
            with tqdm(total=len(recordings)) as self.pbar:
                while offset <= len(recordings):
                    self.process_recordings(recordings[offset:offset+self.BATCH_SIZE])
                    offset += self.BATCH_SIZE
        else:
            while offset <= len(recordings):
                self.process_recordings(recordings[offset:offset+self.BATCH_SIZE])
                offset += self.BATCH_SIZE

    def process_recordings(self, recordings):
        """
            This function carries out the actual lookup of the metadata and inserting the
            popularity and tags into the DB for the given chunk of recordings.
        """

        args = []
        mbid_to_recording = {}
        for rec in recordings:
            mbid_to_recording[rec.mbid] = rec
            args.append({"recording_mbid": rec.mbid})

        while True:
            r = requests.post("https://labs.api.listenbrainz.org/bulk-tag-lookup/json", json=args)
            if r.status_code == 429:
                sleep(2)
                continue

            if r.status_code != 200:
                logger.info("Fail: %d %s" % (r.status_code, r.text))
                return False

            break

        recording_pop = {}
        recording_tags = defaultdict(lambda: defaultdict(list))
        tags = set()
        for row in r.json():
            mbid = str(row["recording_mbid"])
            recording_pop[mbid] = row["percent"]
            recording_tags[mbid][row["source"]].append(row["tag"])
            tags.add(row["tag"])

        if not self.quiet:
            self.pbar.update(len(recordings))

        with db.atomic():

            # This DB code is pretty messy -- things I take for granted with Postgres are not
            # available in SQLite or the PeeWee ORM. But, this might be ok, since we're not
            # updating millions of rows constantly.

            # First update recording_metadata table
            for mbid in set(recording_pop):
                recording = mbid_to_recording[mbid]
                if recording.metadata_id is None:
                    recording_metadata = RecordingMetadata.create(recording=recording.id,
                                                                  popularity=recording_pop[mbid],
                                                                  last_updated=datetime.datetime.now())
                    recording_metadata.save()
                else:
                    recording_metadata = RecordingMetadata.replace(id=recording.metadata_id,
                                                                   recording=recording.id,
                                                                   popularity=recording_pop[mbid],
                                                                   last_updated=datetime.datetime.now())

                    recording_metadata.execute()

            # Next delete recording_tags
            RecordingTag.delete().where(
                RecordingTag.recording_id.in_(
                    Recording.select(Recording.id).where(
                        Recording.recording_mbid.in_(set(recording_tags))

                    )
                )
            ).execute()
            # This is the better way to insert the tags into the DB, but on some installations
            # of Sqlite/Python the UPSERT is not supported. Once it is widely supported,
            # remove the section below and uncomment this.
            #tag_ids = {}
            #for tag in tags:
            #    cursor = db.execute_sql("""INSERT INTO tag (name)
            #                                    VALUES (?)
            #                 ON CONFLICT DO UPDATE SET name = ? RETURNING id""", (tag,tag))
            #    row = cursor.fetchone()
            #    tag_ids[tag] = row[0]

            # insert new recording tags
            for tag in tags:
                db.execute_sql("""INSERT OR IGNORE INTO tag (name) VALUES (?)""", (tag,))

            tag_str = ",".join([ "'%s'" % t.replace("'", "''") for t in tags])
            cursor = db.execute_sql("""SELECT id, name FROM tag WHERE name IN (%s)""" % tag_str)
            tag_ids = {row[1]: row[0] for row in cursor.fetchall()}

            # insert recording_tag rows
            with db.atomic():
                now = datetime.datetime.now()
                for row in r.json():
                    recording = mbid_to_recording[row["recording_mbid"]]
                    db.execute_sql("""INSERT INTO recording_tag (recording_id, tag_id, entity, last_updated)
                                       VALUES (?, ?, ?, ?)""", (recording.id, tag_ids[row["tag"]], row["source"], now))

        return True



import json
import unittest
import unittest.mock

import troi
import troi.musicbrainz.recording_lookup

return_json = {
    "1234a7ae-2af2-4291-aa84-bd0bafe291a1": {
        "artist": {
            "artist_credit_id":
            65,
            "artists": [{
                "area": "United Kingdom",
                "artist_mbid": "8f6bd1e4-fbe1-4f50-aa9b-94c450ec0f11",
                "begin_year": 1991,
                "join_phrase": "",
                "name": "Portishead",
                "rels": {
                    "free streaming": "https://www.deezer.com/artist/1069",
                    "lyrics": "https://muzikum.eu/en/122-6105/portishead/lyrics.html",
                    "official homepage": "http://www.portishead.co.uk/",
                    "purchase for download": "https://www.junodownload.com/artists/Portishead/releases/",
                    "social network": "https://www.facebook.com/portishead",
                    "streaming": "https://tidal.com/artist/27441",
                    "wikidata": "https://www.wikidata.org/wiki/Q191352",
                    "youtube": "https://www.youtube.com/user/portishead1002"
                },
                "type": "Group"
            }],
            "name":
            "Portishead"
        },
        "recording": {
            "length":
            253960,
            "name":
            "Sour Times",
            "rels": [{
                "artist_mbid": "5adcb9d9-5ea2-428d-af46-ef626966e106",
                "artist_name": "Beth Gibbons",
                "instrument": "lead vocals",
                "type": "vocal"
            }, {
                "artist_mbid": "eece77f9-9c15-4583-87f8-37d1b7b47e36",
                "artist_name": "Neil Solman",
                "instrument": "Rhodes piano",
                "type": "instrument"
            }, {
                "artist_mbid": "619b1116-740e-42e0-bdfe-96af274f79f7",
                "artist_name": "Adrian Utley",
                "instrument": "guitar",
                "type": "instrument"
            }, {
                "artist_mbid": "eece77f9-9c15-4583-87f8-37d1b7b47e36",
                "artist_name": "Neil Solman",
                "instrument": "Hammond organ",
                "type": "instrument"
            }]
        },
        "release": {
            "album_artist_name": "Portishead",
            "caa_id": 829521842,
            "caa_release_mbid": "76df3287-6cda-33eb-8e9a-044b5e15ffdd",
            "mbid": "76df3287-6cda-33eb-8e9a-044b5e15ffdd",
            "name": "Dummy",
            "release_group_mbid": "48140466-cff6-3222-bd55-63c27e43190d",
            "year": 1994
        },
        "tag": {
            "artist": [{
                "artist_mbid": "8f6bd1e4-fbe1-4f50-aa9b-94c450ec0f11",
                "count": 9,
                "genre_mbid": "cc38aba3-48ed-439a-83b9-f81a34a66598",
                "tag": "downtempo"
            }, {
                "artist_mbid": "8f6bd1e4-fbe1-4f50-aa9b-94c450ec0f11",
                "count": 12,
                "genre_mbid": "45eb1d9c-588c-4dc8-9394-a14b7c8f02bc",
                "tag": "trip hop"
            }, {
                "artist_mbid": "8f6bd1e4-fbe1-4f50-aa9b-94c450ec0f11",
                "count": 6,
                "genre_mbid": "89255676-1f14-4dd8-bbad-fca839d6aff4",
                "tag": "electronic"
            }, {
                "artist_mbid": "8f6bd1e4-fbe1-4f50-aa9b-94c450ec0f11",
                "count": 7,
                "tag": "trip-hop"
            }, {
                "artist_mbid": "8f6bd1e4-fbe1-4f50-aa9b-94c450ec0f11",
                "count": 1,
                "tag": "the lost generation"
            }, {
                "artist_mbid": "8f6bd1e4-fbe1-4f50-aa9b-94c450ec0f11",
                "count": 1,
                "genre_mbid": "6e2e809f-8c54-4e0f-aca0-0642771ab3cf",
                "tag": "electro-industrial"
            }, {
                "artist_mbid": "8f6bd1e4-fbe1-4f50-aa9b-94c450ec0f11",
                "count": 4,
                "genre_mbid": "65c97e89-b42b-45c2-a70e-0eca1b8f0ff7",
                "tag": "experimental rock"
            }, {
                "artist_mbid": "8f6bd1e4-fbe1-4f50-aa9b-94c450ec0f11",
                "count": 1,
                "genre_mbid": "ec5a14c7-7793-46dc-b858-470183eb63f7",
                "tag": "folktronica"
            }, {
                "artist_mbid": "8f6bd1e4-fbe1-4f50-aa9b-94c450ec0f11",
                "count": 3,
                "tag": "british"
            }, {
                "artist_mbid": "8f6bd1e4-fbe1-4f50-aa9b-94c450ec0f11",
                "count": 2,
                "genre_mbid": "ba318056-9ddf-46cd-8b95-61fc993b962d",
                "tag": "krautrock"
            }],
            "recording": [{
                "count": 3,
                "genre_mbid": "89255676-1f14-4dd8-bbad-fca839d6aff4",
                "tag": "electronic"
            }, {
                "count": 1,
                "genre_mbid": "a2782cb6-1cd0-477c-a61d-b3f8b42dd1b3",
                "tag": "house"
            }, {
                "count": 1,
                "genre_mbid": "7dc2b20f-3953-4874-b9bf-41b8ba06d20c",
                "tag": "acid jazz"
            }, {
                "count": 1,
                "genre_mbid": "ceeaa283-5d7b-4202-8d1d-e25d116b2a18",
                "tag": "alternative rock"
            }, {
                "count": 3,
                "genre_mbid": "cc38aba3-48ed-439a-83b9-f81a34a66598",
                "tag": "downtempo"
            }, {
                "count": 10,
                "genre_mbid": "45eb1d9c-588c-4dc8-9394-a14b7c8f02bc",
                "tag": "trip hop"
            }, {
                "count": 1,
                "genre_mbid": "bab69b07-8bb9-4415-b666-c37609cef80b",
                "tag": "club"
            }, {
                "count": 1,
                "genre_mbid": "e5bba957-8c91-496a-a675-c6d0c6b51c33",
                "tag": "dance"
            }, {
                "count": 4,
                "tag": "trip-hop"
            }],
            "release_group": [{
                "count": 7,
                "tag": "trip-hop"
            }, {
                "count": 1,
                "tag": "british"
            }, {
                "count": 2,
                "tag": "britannique"
            }, {
                "count": 5,
                "genre_mbid": "89255676-1f14-4dd8-bbad-fca839d6aff4",
                "tag": "electronic"
            }, {
                "count": 6,
                "genre_mbid": "cc38aba3-48ed-439a-83b9-f81a34a66598",
                "tag": "downtempo"
            }, {
                "count": 2,
                "genre_mbid": "7dc2b20f-3953-4874-b9bf-41b8ba06d20c",
                "tag": "acid jazz"
            }, {
                "count": 1,
                "genre_mbid": "fd7b5582-e69c-40ad-90b1-d31db33be612",
                "tag": "dark jazz"
            }, {
                "count": 1,
                "genre_mbid": "ec5a14c7-7793-46dc-b858-470183eb63f7",
                "tag": "folktronica"
            }, {
                "count": 1,
                "genre_mbid": "ceeaa283-5d7b-4202-8d1d-e25d116b2a18",
                "tag": "alternative rock"
            }, {
                "count": 1,
                "genre_mbid": "da52c9d1-73f2-456e-a0d0-94bd854cb338",
                "tag": "illbient"
            }, {
                "count": 1,
                "tag": "electronica dance"
            }, {
                "count": 17,
                "genre_mbid": "45eb1d9c-588c-4dc8-9394-a14b7c8f02bc",
                "tag": "trip hop"
            }]
        }
    },
    "ec5b8aa9-7483-4791-a185-1f599a0cdc35": {
        "artist": {
            "artist_credit_id":
            11,
            "artists": [{
                "area": "United States",
                "artist_mbid": "31810c40-932a-4f2d-8cfd-17849844e2a6",
                "begin_year": 1993,
                "join_phrase": "",
                "name": "Squirrel Nut Zippers",
                "rels": {
                    "free streaming": "https://www.deezer.com/artist/7434",
                    "lyrics": "https://genius.com/artists/Squirrel-nut-zippers",
                    "official homepage": "http://snzippers.com/",
                    "social network": "https://www.facebook.com/SNZippers",
                    "streaming": "https://tidal.com/artist/3533610",
                    "wikidata": "https://www.wikidata.org/wiki/Q1817585",
                    "youtube": "https://www.youtube.com/user/snztv"
                },
                "type": "Group"
            }],
            "name":
            "Squirrel Nut Zippers"
        },
        "recording": {
            "length": 275333,
            "name": "Blue Angel",
            "rels": []
        },
        "release": {
            "album_artist_name": "Squirrel Nut Zippers",
            "caa_id": 8413711733,
            "caa_release_mbid": "70fc4df9-1a86-4357-aac7-0694d4248aed",
            "mbid": "70fc4df9-1a86-4357-aac7-0694d4248aed",
            "name": "Hot",
            "release_group_mbid": "c6fe6a2b-0ed6-3d2c-b9ce-ddd5421a3452",
            "year": 1996
        },
        "tag": {
            "artist": [{
                "artist_mbid": "31810c40-932a-4f2d-8cfd-17849844e2a6",
                "count": 3,
                "genre_mbid": "a715278f-1580-409f-8078-4ffbc800e08b",
                "tag": "jazz"
            }, {
                "artist_mbid": "31810c40-932a-4f2d-8cfd-17849844e2a6",
                "count": 1,
                "tag": "american"
            }, {
                "artist_mbid": "31810c40-932a-4f2d-8cfd-17849844e2a6",
                "count": 2,
                "genre_mbid": "6162d71d-6307-4021-8c50-a9524444f645",
                "tag": "swing revival"
            }, {
                "artist_mbid": "31810c40-932a-4f2d-8cfd-17849844e2a6",
                "count": 1,
                "genre_mbid": "063d0461-1e12-4246-87bf-69c5ca611cc2",
                "tag": "swing"
            }],
            "recording": [{
                "count": 1,
                "genre_mbid": "e7bb9a21-9556-4b85-bc78-5f69a3d0577d",
                "tag": "big band"
            }, {
                "count": 2,
                "genre_mbid": "6162d71d-6307-4021-8c50-a9524444f645",
                "tag": "swing revival"
            }, {
                "count": 6,
                "genre_mbid": "a715278f-1580-409f-8078-4ffbc800e08b",
                "tag": "jazz"
            }, {
                "count": 1,
                "genre_mbid": "d017dd87-704b-4ad4-8b67-39fb401b8339",
                "tag": "ragtime"
            }, {
                "count": 4,
                "genre_mbid": "063d0461-1e12-4246-87bf-69c5ca611cc2",
                "tag": "swing"
            }],
            "release_group": [{
                "count": 1,
                "genre_mbid": "d017dd87-704b-4ad4-8b67-39fb401b8339",
                "tag": "ragtime"
            }, {
                "count": 4,
                "genre_mbid": "a715278f-1580-409f-8078-4ffbc800e08b",
                "tag": "jazz"
            }, {
                "count": 2,
                "genre_mbid": "063d0461-1e12-4246-87bf-69c5ca611cc2",
                "tag": "swing"
            }, {
                "count": 1,
                "genre_mbid": "6162d71d-6307-4021-8c50-a9524444f645",
                "tag": "swing revival"
            }]
        }
    }
}


class TestRecordingLookup(unittest.TestCase):

    @unittest.mock.patch('requests.post')
    def test_read(self, req):

        mock = unittest.mock.MagicMock()
        mock.status_code = 200
        mock.text = json.dumps(return_json)
        req.return_value = mock
        e = troi.musicbrainz.recording_lookup.RecordingLookupElement(lookup_tags=True)

        inputs = [
            troi.Recording(mbid="1234a7ae-2af2-4291-aa84-bd0bafe291a1"),
            troi.Recording(mbid="ec5b8aa9-7483-4791-a185-1f599a0cdc35")
        ]
        entities = e.read([inputs])
        req.assert_called_with(e.SERVER_URL,
                               json={
                                   'recording_mbids':
                                   ['1234a7ae-2af2-4291-aa84-bd0bafe291a1', 'ec5b8aa9-7483-4791-a185-1f599a0cdc35'],
                                   'inc': 'artist release tag'
                               },
                               headers={})

        assert len(entities) == 2

        assert entities[0].name == "Sour Times"
        assert entities[0].duration == 253960
        assert entities[0].year == 1994
        assert entities[0].artist_credit.name == "Portishead"
        assert entities[0].artist_credit.artist_credit_id == 65
        assert entities[0].mbid == "1234a7ae-2af2-4291-aa84-bd0bafe291a1"
        assert entities[0].release.mbid == "76df3287-6cda-33eb-8e9a-044b5e15ffdd"
        assert entities[0].release.name == "Dummy"

        # Check tags
        assert entities[0].artist_credit.musicbrainz["genre"] == [
            'downtempo', 'trip hop', 'electronic', 'electro-industrial', 'experimental rock', 'folktronica', 'krautrock'
        ]
        assert entities[0].artist_credit.musicbrainz["tag"] == ['trip-hop', 'the lost generation', 'british']
        assert entities[0].release.musicbrainz["genre"] == [
            'electronic', 'downtempo', 'acid jazz', 'dark jazz', 'folktronica', 'alternative rock', 'illbient', 'trip hop'
        ]
        assert entities[0].release.musicbrainz["tag"] == ['trip-hop', 'british', 'britannique', 'electronica dance']
        assert entities[0].musicbrainz["genre"] == [
            'electronic', 'house', 'acid jazz', 'alternative rock', 'downtempo', 'trip hop', 'club', 'dance'
        ]
        assert entities[0].musicbrainz["tag"] == ['trip-hop']

        assert entities[1].name == "Blue Angel"
        assert entities[1].duration == 275333
        assert entities[1].year == 1996
        assert entities[1].artist_credit.name == "Squirrel Nut Zippers"
        assert entities[1].artist_credit.artist_credit_id == 11
        assert entities[1].mbid == "ec5b8aa9-7483-4791-a185-1f599a0cdc35"
        assert entities[1].release.mbid == "70fc4df9-1a86-4357-aac7-0694d4248aed"
        assert entities[1].release.name == "Hot"

        # Check tags
        assert entities[1].artist_credit.musicbrainz["genre"] == ['jazz', 'swing revival', 'swing']
        assert entities[1].artist_credit.musicbrainz["tag"] == ['american']
        assert entities[1].release.musicbrainz["genre"] == ['ragtime', 'jazz', 'swing', 'swing revival']
        assert entities[1].release.musicbrainz["tag"] == []
        assert entities[1].musicbrainz["genre"] == ['big band', 'swing revival', 'jazz', 'ragtime', 'swing']
        assert entities[1].musicbrainz["tag"] == []



import datetime
import logging

from troi import Recording, Playlist, PipelineError

logger = logging.getLogger(__name__)

class PrintRecordingList:
    """
        Print a list of recordings in a sane matter intended to fit on a reasonably sized screen.
        It prints recording name and artist name always, and year, bpm, listen_count or moods
        if they are found in the first recording.
    """

    def __init__(self):
        super().__init__()
        self.print_year = False
        self.print_bpm = False
        self.print_listen_count = False
        self.print_moods = False
        self.print_genre = False
        self.print_latest_listened_at = False
        self.print_ranking = False
        self.print_popularity = False

    def _examine_recording_for_headers(self, recording):
        # Look at the first item and decide which columns to show
        if recording.year is not None:
            self.print_year = True

        if "listen_count" in recording.listenbrainz:
            self.print_listen_count = True

        if "bpm" in recording.acousticbrainz:
            self.print_bpm = True

        if "moods" in recording.acousticbrainz:
            self.print_moods = True

        if "genres" in recording.musicbrainz or "tags" in recording.musicbrainz:
            self.print_genre = True

        if "latest_listened_at" in recording.listenbrainz:
            self.print_latest_listened_at = True

        if recording.ranking:
            self.print_ranking = True
        
        if "popularity" in recording.musicbrainz:
            self.print_popularity = True

    def _print_recording(self, recording, year=False, popularity=False, listen_count=False, bpm=False, moods=False, genre=False):
        """ Print out a recording, formatting it nicely to fit in a reasonably sized window.
            The year, listen_count, bpm, mood and genre arguments here can override the settings
            gleaned from the first recording submitted to this class"""

        if recording.artist_credit is None:
            artist = "[missing]"
        elif recording.artist_credit.name:
            artist = recording.artist_credit.name
        else: 
            artist = "[[ unknown ]]"

        if recording.name is None:
            rec_name = "[[ mbid:%s ]]" % recording.mbid
        else:
            rec_name = recording.name
        rec_mbid = recording.mbid[:5] if recording.mbid else "[[ ]]"

        text = "%-60s %-50s %5s" % (rec_name[:59], artist[:49], rec_mbid)

        if recording.artist_credit is not None:
            if recording.artist_credit.artists is not None:
                text += " %-20s" % ",".join([a.mbid[:5] for a in recording.artist_credit.artists])
            if recording.artist_credit.artist_credit_id is not None:
                text += " %8d" % recording.artist_credit.artist_credit_id

        if self.print_year and recording.year is not None:
            text += " %d" % recording.year
        if self.print_ranking:
            text += " %.3f" % recording.ranking
        if self.print_listen_count or listen_count:
            text += " %4d" % recording.listenbrainz['listen_count']
        if self.print_bpm or bpm:
            text += " %3d" % recording.acousticbrainz['bpm']
        if self.print_popularity or popularity:
            text += " %.1f" % recording.musicbrainz.get('popularity', 0.0)
        if self.print_latest_listened_at:
            if recording.listenbrainz["latest_listened_at"] is None:
                text += " never    "
            else:
                now = datetime.datetime.now()
                td = now - recording.listenbrainz["latest_listened_at"]
                text += " %3d days " % td.days
        if self.print_moods or moods:
            # TODO: make this print more than agg, but given the current state of moods/coverage...
            text = " mood agg %3d" % int(100 * recording.acousticbrainz['moods']["mood_aggressive"])
        if self.print_genre or genre:
            text = " %s" % ",".join(recording.musicbrainz.get("genres", []))
            text = " %s" % ",".join(recording.musicbrainz.get("tags", []))

        logger.info(text)

    def print(self, entity):
        """ Print out a list(Recording) or list(Playlist). """

        if type(entity) == Recording:
            self._examine_recording_for_headers(entity)
            self._print_recording(entity)
            return

        for rec in entity:
            self._examine_recording_for_headers(rec)

        if type(entity) == list and type(entity[0]) == Recording:
            for rec in entity:
                self._print_recording(rec)

        if type(entity) == Playlist:
            for rec in entity.recordings:
                self._print_recording(rec)

        raise PipelineError("You must pass a Recording or list of Recordings or a Playlist to print.")



def get_tag_value(tags, tag, default=None):
    """
        Get the tag value and return the first item in the list or None if that is not possible.
    """

    try:
        t = tags[tag]
    except KeyError:
        return default

    return t[0]


def extract_track_number(track_number):
    """
        Parse the various forms of track number formats and return a simple integer. If nothing
        sensible is found, return None
    """

    if track_number is None:
        return None

    # Used for m4a tags
    if isinstance(track_number, tuple):
        return track_number[0]

    # Used for ID3
    if str(track_number).find("/") != -1:
        track_number, dummy = str(track_number).split("/")
    try:
        return_value = int(track_number)
    except ValueError:
        return_value = None

    return return_value


def make_artist_array(artist_id):
    """
        Given artist id tag data, return a string from the data.
        Accepts: list, string. If something else is passed, cast to str.
    """
    if isinstance(artist_id, str):
        return artist_id

    if isinstance(artist_id, list):
        return ",".join(artist_id)

    return str(artist_id)



from datetime import datetime, timedelta

from troi.patch import Patch
import troi.filters
import troi.listenbrainz.feedback
import troi.listenbrainz.listens
import troi.listenbrainz.recs
import troi.musicbrainz.recording_lookup
from troi import Playlist
from troi.playlist import PlaylistMakerElement

DAYS_OF_RECENT_LISTENS_TO_EXCLUDE = 60  # Exclude tracks listened in last X days from the daily jams playlist
DAILY_JAMS_MIN_RECORDINGS = 25  # the minimum number of recordings we aspire to have in a daily jam, this is not a hard limit
BATCH_SIZE_RECS = 100  # the number of recommendations fetched in 1 go
MAX_RECS_LIMIT = 1000  # the maximum of recommendations available in LB

WEEKLY_JAMS_DESCRIPTION = """<p>The ListenBrainz Weekly Jams playlist features songs that you have listened to before, 
    arranged into a comfortable playlist that doesn't require active listening on your part.</p>

    <p>Generated on Monday morning, in accordance with your timezone settings.</p>
"""

WEEKLY_EXPLORATION_DESCRIPTION = """<p>The ListenBrainz Weekly Exploration playlist helps you discover new music! 
    It may require active listening and skips. The playlist features tracks you haven't heard before, 
    selected by a collaborative filtering algorithm.</p>

    <p>Updated every Monday morning based on your timezone.</p>
"""


class PeriodicJamsPatch(Patch):
    """
       Create either daily-jams, weekly-jams or weekly-exploration with this patch.

       First, fetch the top recommendations. For daily-jams and weekly-jams, filter out the recently listened tracks.
       For weekly-exploration, filter out tracks that have been listened to.

       Then filter out hated tracks and make the playlist.
    """

    JAM_TYPES = ("daily-jams", "weekly-jams", "weekly-exploration")

    def __init__(self, args):
        super().__init__(args)

    @staticmethod
    def inputs():
        """
        Generate a periodic playlist from the ListenBrainz recommended recordings.

        \b
        USER_NAME is a MusicBrainz user name that has an account on ListenBrainz.
        TYPE Must be one of "daily-jams", "weekly-jams" or "weekly-exploration".
        JAM_DATE is the date for which the jam is created (this is needed to account for the fact different timezones
        can be on different dates). Required formatting for the date is 'YYYY-MM-DD'.
        """
        return [{
            "type": "argument",
            "args": ["user_name"]
        }, {
            "type": "argument",
            "args": ["type"],
            "kwargs": {
                "required": False
            }
        }, {
            "type": "argument",
            "args": ["jam_date"],
            "kwargs": {
                "required": False
            }
        }]

    @staticmethod
    def outputs():
        return [Playlist]

    @staticmethod
    def slug():
        return "periodic-jams"

    @staticmethod
    def description():
        return "Generate a periodic playlist from the ListenBrainz recommended recordings."

    def create(self, inputs):
        user_name = inputs['user_name']
        jam_date = inputs.get('jam_date')
        if jam_date is None:
            jam_date = datetime.utcnow().strftime("%Y-%m-%d %a")
        jam_type = inputs.get('type')
        if jam_type is None:
            jam_type = self.JAM_TYPES[0]
        else:
            jam_type = jam_type.lower()
            if jam_type not in self.JAM_TYPES:
                raise RuntimeError("Jam type must be one of %s" % ", ".join(self.JAM_TYPES))

        recs = troi.listenbrainz.recs.UserRecordingRecommendationsElement(user_name,
                                                                          "raw",
                                                                          count=1000,
                                                                          auth_token=inputs.get("token"))

        recent_listens_lookup = troi.listenbrainz.listens.RecentListensTimestampLookup(user_name,
                                                                                       days=2,
                                                                                       auth_token=inputs.get("token"))
        recent_listens_lookup.set_sources(recs)

        if jam_type in ("daily-jams", "weekly-jams"):
            # Remove tracks that have not been listened to before.
            never_listened = troi.filters.NeverListenedFilterElement(keep_unlistened_if_empty=True)
            never_listened.set_sources(recent_listens_lookup)
            if jam_type == "daily-jams":
                jam_name = "Daily Jams"
            else:
                jam_name = "Weekly Jams"
                jam_date = "week of " + jam_date
            self.local_storage["jam_name"] = jam_name

            latest_filter = troi.filters.LatestListenedAtFilterElement(DAYS_OF_RECENT_LISTENS_TO_EXCLUDE,
                                                                       keep_unlistened_if_empty=True)
            latest_filter.set_sources(never_listened)
        elif jam_type == "weekly-exploration":
            # Remove tracks that have been listened to before.
            never_listened = troi.filters.NeverListenedFilterElement(remove_unlistened=False)
            never_listened.set_sources(recent_listens_lookup)
            jam_name = "Weekly Exploration"
            jam_date = "week of " + jam_date
            self.local_storage["jam_name"] = jam_name

            latest_filter = troi.filters.LatestListenedAtFilterElement(DAYS_OF_RECENT_LISTENS_TO_EXCLUDE)
            latest_filter.set_sources(never_listened)

        else:
            raise RuntimeError("someone goofed up!")

        feedback_lookup = troi.listenbrainz.feedback.ListensFeedbackLookup(user_name, auth_token=inputs.get("token"))
        feedback_lookup.set_sources(latest_filter)

        recs_lookup = troi.musicbrainz.recording_lookup.RecordingLookupElement(auth_token=inputs.get("token"))
        recs_lookup.set_sources(feedback_lookup)

        hate_filter = troi.filters.HatedRecordingsFilterElement()
        hate_filter.set_sources(recs_lookup)

        pl_maker = PlaylistMakerElement(name="%s for %s, %s" % (jam_name, user_name, jam_date),
                                        desc=None,
                                        patch_slug=jam_type,
                                        max_num_recordings=50,
                                        max_artist_occurrence=2,
                                        shuffle=True,
                                        expires_at=datetime.utcnow() + timedelta(weeks=2),
                                        is_april_first=(jam_date[5:10] == "04-01"))
        pl_maker.set_sources(hate_filter)

        return pl_maker

    def post_process(self):
        """ 
            Take the information saved in local_storage and create proper playlist names and descriptions.
        """

        jam_name = self.local_storage["jam_name"]
        if self.local_storage["latest_listened_was_empty"]:
            desc = f"This {jam_name} playlist includes tracks you've recently heard, as we don't have enough listens to create a better playlist."
        else:
            desc = f"We made this {jam_name} playlist from your recommended tracks to create a comfortable playlist of music you've not listened to recently."
            
        self.local_storage["_playlist_desc"] = desc



import datetime
from enum import Enum

from peewee import *
from troi.content_resolver.model.database import db


class FileIdType(Enum):
    FILE_PATH = 0
    SUBSONIC_ID = 1


class FileIdTypeField(IntegerField):
    """
    Enum for file id type
    """

    def __init__(self, *args, **kwargs) -> None:
        super(IntegerField, self).__init__(*args, **kwargs)

    def db_value(self, file_id_type_enum):
        return file_id_type_enum.value

    def python_value(self, db_file_id_type):
        return FileIdType(int(db_file_id_type))


class Recording(Model):
    """
    Basic metadata information about a recording on disk (a track).
    """

    class Meta:
        database = db
        indexes = (
            # create a unique on (file_id, file_id_type)
            (('file_id', 'file_id_type'), True),
        )

    id = AutoField()
    file_id = TextField(null=False, index=True)
    file_id_type = FileIdTypeField(null=False, index=True)
    mtime = TimestampField(null=False)

    artist_name = TextField(null=True)
    release_name = TextField(null=True)
    recording_name = TextField(null=True)

    # Not using the UUIDField here, since it annoyingly removes '-' from the UUID.
    recording_mbid = TextField(null=True, index=True)
    artist_mbid = TextField(null=True, index=True)
    release_mbid = TextField(null=True, index=True)

    duration = IntegerField(null=True)
    track_num = IntegerField(null=True)
    disc_num = IntegerField(null=True)

    def __repr__(self):
        return "<Recording('%s','%s')>" % (self.recording_mbid or "", self.recording_name)


class RecordingMetadata(Model):
    """
    Additional metadata for recordings: popularity. In future additional fields
    like release date and release country could be added to this table.
    """

    class Meta:
        database = db
        table_name = "recording_metadata"

    id = AutoField()
    recording = ForeignKeyField(Recording, backref="metadata")

    popularity = FloatField()
    last_updated = DateTimeField(null=False, default=datetime.datetime.now)

    def __repr__(self):
        return "<RecordingMetadata('%d','%.3f')>" % (self.recording or 0, self.popularity)



import os
from collections import defaultdict
import datetime
import sys
from time import sleep

import peewee
import requests

from troi.content_resolver.model.database import db
from troi.content_resolver.model.recording import Recording, RecordingMetadata
from troi.content_resolver.utils import select_recordings_on_popularity
from troi.recording_search_service import RecordingSearchByArtistService
from troi.plist import plist

OVERHYPED_SIMILAR_ARTISTS = [
    "b10bbbfc-cf9e-42e0-be17-e2c3e1d2600d",  # The Beatles
    "83d91898-7763-47d7-b03b-b92132375c47",  # Pink Floyd
    "a74b1b7f-71a5-4011-9441-d0b5e4122711",  # Radiohead
    "8bfac288-ccc5-448d-9573-c33ea2aa5c30",  # Red Hot Chili Peppers
    "9c9f1380-2516-4fc9-a3e6-f9f61941d090",  # Muse
    "cc197bad-dc9c-440d-a5b5-d52ba2e14234",  # Coldplay
    "65f4f0c5-ef9e-490c-aee3-909e7ae6b2ab",  # Metallica
    "5b11f4ce-a62d-471e-81fc-a69a8278c7da",  # Nirvana
    "f59c5520-5f46-4d2c-b2c4-822eabf53419",  # Linkin Park
    "cc0b7089-c08d-4c10-b6b0-873582c17fd6",  # System of a Down
    "ebfc1398-8d96-47e3-82c3-f782abcdb13d",  # Beach boys
]

class LocalRecordingSearchByArtistService(RecordingSearchByArtistService):
    '''
    Given the local database, search for artists that meet given tag criteria
    '''

    def __init__(self):
        RecordingSearchByArtistService.__init__(self)

    def get_similar_artists(self, artist_mbid):
        """ Fetch similar artists, given an artist_mbid. Returns a sored plist of artists. """

        while True:
            r = requests.post("https://labs.api.listenbrainz.org/similar-artists/json",
                              json=[{
                                  'artist_mbids':
                                  [artist_mbid],
                                  'algorithm':
                                  "session_based_days_7500_session_300_contribution_5_threshold_10_limit_100_filter_True_skip_30"
                              }])
            if r.status_code == 429:
                sleep(2)
                continue
            if r.status_code != 200:
                raise RuntimeError(f"Cannot fetch similar artists: {r.status_code} ({r.text})")

            break

        artists = r.json()

        # Knock down super hyped artists
        for artist in artists:
            if artist["artist_mbid"] in OVERHYPED_SIMILAR_ARTISTS:
                artist["score"] /= 3  # Chop!

        return plist(sorted(artists, key=lambda a: a["score"], reverse=True))

    def search(self, mode, artist_mbid, pop_begin, pop_end, max_recordings_per_artist, max_similar_artists):

        """
        Perform an artist search. Parameters:

        mode: the mode used for this artist search
        pop_begin: if many recordings match the above parameters, return only
                       recordings that have a minimum popularity percent score
                       of pop_begin.
        pop_end: if many recordings match the above parameters, return only
                     recordings that have a maximum popularity percent score
                     of pop_end.
        max_recordings_per_artist: The number of recordings to collect for each artist.
        max_similar_artists: The maximum number of similar artists to select.

        If only few recordings match, the pop_begin and pop_end are ignored.
        """

        similar_artists = self.get_similar_artists(artist_mbid)
        query = """SELECT popularity
                        , recording_mbid
                        , artist_mbid
                        , file_id
                        , file_id_type
                     FROM recording
                     JOIN recording_metadata
                       ON recording.id = recording_metadata.recording_id
                    WHERE artist_mbid in (%s)
                 ORDER BY artist_mbid
                        , popularity"""

        artist_mbids = [artist["artist_mbid"] for artist in similar_artists]
        placeholders = ",".join(("?", ) * len(similar_artists))
        cursor = db.execute_sql(query % placeholders, params=artist_mbids)

        artists = defaultdict(list)
        for rec in cursor.fetchall():
            artists[rec[2]].append({
                "popularity": rec[0],
                "recording_mbid": rec[1],
                "artist_mbid": rec[2],
                "file_id": rec[3],
                "file_id_type": rec[4]
            })

        for artist in artists:
            artists[artist] = select_recordings_on_popularity(artists[artist], pop_begin, pop_end, max_recordings_per_artist)

        return artists, []



import logging

from troi import Playlist
from troi.playlist import PlaylistElement
from troi.patches.lb_radio import LBRadioPatch

from troi.content_resolver.tag_search import LocalRecordingSearchByTagService
from troi.content_resolver.artist_search import LocalRecordingSearchByArtistService
from troi.content_resolver.content_resolver import ContentResolver

logger = logging.getLogger(__name__)


class ListenBrainzRadioLocal:
    '''
       Generate local playlists against a music collection available via subsonic.
    '''

    def __init__(self, quiet):
        self.quiet = quiet

    def generate(self, mode, prompt, match_threshold):
        """
           Generate a playlist given the mode and prompt. Optional match_threshold, a value from
           0 to 1.0 allows the use to control how well local resolution tracks must match before
           being considered a match.

           Returns a troi playlist object.
        """

        patch = LBRadioPatch({"mode": mode, "prompt": prompt, "quiet": self.quiet, "min_recordings": 1})
        patch.register_service(LocalRecordingSearchByTagService())
        patch.register_service(LocalRecordingSearchByArtistService())

        # Now generate the playlist
        try:
            playlist = patch.generate_playlist()
        except RuntimeError as err:
            logger.info(f"LB Radio generation failed: {err}")
            return None

        if playlist is None:
            return playlist

        # Resolve any tracks that have not been resolved to a subsonic_id or a local file
        self.resolve_playlist(match_threshold, playlist)

        return playlist

    def resolve_playlist(self, match_threshold, playlist):
        """ Attempt to resolve any tracks without local ids to local ids """

        # Find recordings that are missing local ids
        recordings = []
        for recording in playlist.playlists[0].recordings:
            if "subsonic_id" in recording.musicbrainz or "filename" in recording.musicbrainz:
                continue

            recordings.append(recording)

        if not recordings:
            return

        # Use the content resolver to resolve the recordings in situ
        cr = ContentResolver(self.quiet)
        pe = PlaylistElement()
        pe.playlists = [ Playlist(recordings=recordings) ]
        cr.resolve_playlist(match_threshold, pe)

        # Now filter out the tracks that were not matched
        filtered = []
        for rec in playlist.playlists[0].recordings:
            if "subsonic_id" in rec.musicbrainz or "filename" in rec.musicbrainz:
                filtered.append(rec)

        playlist.playlists[0].recordings = filtered



from collections import defaultdict
import datetime
from operator import itemgetter
from random import shuffle

import troi
from troi import PipelineError, Recording, Playlist


class ArtistCreditFilterElement(troi.Element):
    '''
        Remove recordings if they do or do not belong to a given list of artists.

        :param artist_credit_ids: A list of artist_credit_ids to remove/keep
        :param include: If true, include all tracks with the given artist_credit_id, otherwise remove them.
    '''

    def __init__(self, artist_credit_ids, include=False):
        '''
            Filter a list of Recordings based on their artist_credit_id.
            The default behaviour is to exclude the artists given, but if
            include=True then only the artists in the given list pass
            the filter. Throws PipelineError is not all recordings have
            artist_credit_ids set.
        '''
        super().__init__()
        self.artist_credit_ids = artist_credit_ids
        self.include = include

    @staticmethod
    def inputs():
        return [Recording, Playlist]

    @staticmethod
    def outputs():
        return [Recording, Playlist]

    def read(self, inputs):

        recordings = inputs[0]
        ac_index = {}
        for ac in self.artist_credit_ids:
            try:
                ac_index[ac] = 1
            except KeyError:
                raise PipelineError(self.__name__ + " needs to have all input recordings to have artist.artist_credit_id defined!")

        results = []
        for r in recordings:
            if not r.artist_credit or not r.artist_credit.artist_credit_id:
                continue

            if self.include:
                if r.artist_credit.artist_credit_id in ac_index:
                    results.append(r)
            else:
                if r.artist_credit.artist_credit_id not in ac_index:
                    results.append(r)

        return results


class ArtistCreditLimiterElement(troi.Element):
    '''
        This element examines there passed in recordings and if the count of
        recordings by any one artists exceeds the given limit, excessive recordigns
        are removed. If the flag exclude_lower_ranked is True, and each recording
        has a "ranked" key in the musicbrainz dict, then the lowest
        ranked recordings are removed, otherwise the highest ranked recordings
        are removed. Throws PipelineError is not all recordings have
        artist_credit_ids set.

        :param count: The number of duplicate aritst_credits to allow in the output
        :param exclude_lower_ranked: Remove the lower ranked duplicates, if rankings are present.
    '''

    def __init__(self, count=2, exclude_lower_ranked=True):
        troi.Element.__init__(self)
        self.count = count
        self.exclude_lower_ranked = exclude_lower_ranked

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def _filter(self, recordings):
        """
            Carry out the actual artist limiting.
        """

        ac_index = defaultdict(list)
        all_have_rankings = True
        for rec in recordings:
            try:
                ac_index[rec.artist_credit.artist_credit_id].append((rec.mbid, rec.ranking))
                if rec.ranking is None:
                    all_have_rankings = False
            except KeyError:
                raise PipelineError(self.__name__ + " needs to have all input recordings to have artist.artist_credit_id defined!")

        for key in ac_index:
            if all_have_rankings:
                ac_index[key] = sorted(ac_index[key], key=itemgetter(1), reverse=self.exclude_lower_ranked)
            else:
                shuffle(ac_index[key])
            ac_index[key] = ac_index[key][:self.count]

        pass_recs = []
        for key in ac_index:
            for mbid, ranking in ac_index[key]:
                pass_recs.append(mbid)

        results = []
        for r in recordings:
            if r.mbid in pass_recs:
                results.append(r)

        return results


    def read(self, inputs):
        """
            Determine if recordings or playlists are passed in and call the internal _filter
            function accordingly.
        """

        outputs = []
        for input in inputs:
            if isinstance(input[0], Recording):
                return self._filter(input)
            elif isinstance(input[0], Playlist):
                for playlist in input:
                    playlist.recordings = self._filter(playlist.recordings)
                    outputs.append(playlist)
            else:
                raise PipelineError("ArtistCreditLimiter passed incorrect input types.")

        return outputs


class DuplicateRecordingMBIDFilterElement(troi.Element):
    """
        This Element takes a list of recordings and removes any duplicate recordings
        based on the recording's MBID, preserving the input order.
    """

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs):
        recordings = inputs[0]
        output = []
        seen = set()
        for rec in recordings:
            if rec.mbid not in seen:
                seen.add(rec.mbid)
                output.append(rec)

        return output


class DuplicateRecordingArtistCreditFilterElement(troi.Element):
    """
        This Element takes a list of recordings and removes any duplicate recordings
        based on the recording's name and artist_credit name, preserving the input order.
    """

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs):
        recordings = inputs[0]
        index = {}
        for rec in recordings:
            if rec.name is None or rec.name == "" or rec.artist is None or rec.artist.name is None or rec.artist.name == "":
                continue

            k = rec.name + rec.artist.name
            if k not in index:
                index[k] = rec

        return [ index[k] for k in index ]  


class ConsecutiveRecordingFilterElement(troi.Element):
    """This Element takes a list of recordings and removes consecutive duplicate recordings
    based on the recording's MBID

    For example, a sequence A, A, A, B, B, A, C will be reduced to A, B, A, C
    """

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs):
        recordings = inputs[0]
        output = []
        last_mbid = None
        for rec in recordings:
            if rec.mbid != last_mbid:
                output.append(rec)
            last_mbid = rec.mbid

        return output


class EmptyRecordingFilterElement(troi.Element):
    """This Element takes a list of recordings and removes ones that have an empty name
    or artist.
    """

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, input):
        recordings = inputs[0]
        output = []
        for rec in recordings:
            if rec.name is not None and (rec.artist and rec.artist.name is not None) and rec.mbid is not None:
                output.append(rec)

        return output


class YearRangeFilterElement(troi.Element):
    '''
        Filter a list of Recordings based on their year -- the year must be between
        start_year and end_year, otherwise the recording will be filtered out. If no
        end_year is given, keep (or reject in case of inverse) tracks greater or equal
        to start_year. If inverse=True, then keep all Recordings that do no fit into the
        given year range.

        :param start_year: The full start year to filter (inclusive).
        :param end_year: The full end year to filter (inclusive).
        :param inverse: If inverse is True, exclude everything in the year range.
    '''

    def __init__(self, start_year, end_year=None, inverse=False):
        troi.Element.__init__(self)
        self.start_year = start_year
        self.end_year = end_year
        self.inverse = inverse

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs):

        recordings = inputs[0]

        results = []
        for r in recordings:
            if not r.year:
                continue

            if self.inverse:
                if r.year < self.start_year:
                    results.append(r)
                elif self.end_year and r.year > self.end_year:
                    results.append(r)
            else:
                if r.year >= self.start_year:
                    if not self.end_year:
                        results.append(r)
                    elif r.year <= self.end_year:
                        results.append(r)

        return results


class GenreFilterElement(troi.Element):
    '''
        Keep recorindgs that have at least one genre in commong
        from the list passed in when this class is created.

        :param genre_list: A list of genre trags to filter out.
    '''

    def __init__(self, genre_list):
        troi.Element.__init__(self)
        self.genre_list = genre_list

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs):

        recordings = inputs[0]

        results = []
        for r in recordings:
            if "tags" not in r.musicbrainz:
                continue

            for genre in self.genre_list:
                if genre in r.musicbrainz["tags"]:
                    results.append(r)
                    break

        return results


class LatestListenedAtFilterElement(troi.Element):
    '''
        Filter the recordings according to latest_listened_at field in the lb metadata. 
        If that field is None, treat it as if the user hasn't listened to this track
        recently or at all and keep the track in the list.

        :param min_number_of_days: The number of tracks that must have passed for a track to be kept.
    '''

    MIN_ITEMS_REQUIRED = 5

    def __init__(self, min_number_of_days=14, keep_unlistened_if_empty=False):
        troi.Element.__init__(self)
        self.min_number_of_days = min_number_of_days
        self.keep_unlistened_if_empty = keep_unlistened_if_empty

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs):

        recordings = inputs[0]

        results = []
        now = datetime.datetime.now()
        for r in recordings:
            if "latest_listened_at" in r.listenbrainz and r.listenbrainz["latest_listened_at"] is not None:
                td = now - r.listenbrainz["latest_listened_at"]
                if td.days > self.min_number_of_days:
                    results.append(r)
            else:
                results.append(r)

        if len(results) < self.MIN_ITEMS_REQUIRED and self.keep_unlistened_if_empty:
            self.local_storage["latest_listened_was_empty"] = True
            return recordings

        self.local_storage["latest_listened_was_empty"] = False

        return results


class NeverListenedFilterElement(troi.Element):
    '''
        Remove/keep only recordings if they have not/been listened to.
    '''

    MIN_ITEMS_REQUIRED = 5

    def __init__(self, remove_unlistened=True, keep_unlistened_if_empty=False):
        '''
            Filter the recordings according to latest_listened_at field in the lb metadata. 
            If that field is None, treat it as if the user hasn't listened to this track
            and remove it, or keep it, based on the remove_unlistened parameter.
        '''
        troi.Element.__init__(self)
        self.remove_unlistened = remove_unlistened
        self.keep_unlistened_if_empty = keep_unlistened_if_empty

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs):

        recordings = inputs[0]

        results = []
        for r in recordings:
            # has this track never been listened to before?
            if "latest_listened_at" in r.listenbrainz and r.listenbrainz["latest_listened_at"]:
                if self.remove_unlistened:
                    results.append(r)
                else:
                    continue
            else:
                if self.remove_unlistened:
                    continue
                else:
                    results.append(r)

        if len(results) < self.MIN_ITEMS_REQUIRED and self.keep_unlistened_if_empty:
            self.local_storage["never_listened_was_empty"] = True
            return recordings

        self.local_storage["never_listened_was_empty"] = False
        return results


class HatedRecordingsFilterElement(troi.Element):
    """ Remove recordings that have been hated by the user """

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs):
        results = []
        for r in inputs[0]:
            score = r.listenbrainz.get("score", 0)
            if score < 0:
                continue
            results.append(r)
        return results



import copy
from itertools import zip_longest

import troi


def is_homogeneous(entities):
    """ Check to see if all the items in a list of of the same type """

    if not entities:
        return True

    type_set = set()
    for e in entities:
        type_set.add(type(e))

    return len(type_set) == 1


def _ensure_conformity(entities_0, entities_1):
    """ Check that both entity lists are homogenous and of the same type. """

    if not is_homogeneous(entities_0):
        raise TypeError("entities_0 list not homogenous")

    if not is_homogeneous(entities_1):
        raise TypeError("entities_1 list not homogenous")

    if entities_0 and entities_1 and type(entities_0[0]) != type(entities_1[0]):
        raise TypeError("entities_0 and entities_1 must both be homogenous with all the same type")

    return True


def _check_key_for_set_op(entities, key):
    """ Check the key based on which set operation is to be performed is valid for given entity """
    if isinstance(entities[0], troi.Artist):
        if key not in ['mbid', 'name']:
            raise ValueError("key must be one of mbid, name.")
    if isinstance(entities[0], troi.ArtistCredit):
        if key not in ['name', 'artists', 'artist_credit_id']:
            raise ValueError("key must be one of name, artists or artist_credit_id.")
    elif isinstance(entities[0], troi.Recording):
        if key not in ['mbid', 'msid', 'name']:
            raise ValueError("key must be one of mbid, msid or name.")
    else:
        if key not in ['mbid', 'name']:
            raise ValueError("key must be one of mbid or name.")


class UniqueElement(troi.Element):
    """
        Make this passed list of entities unique base on the passed key
        (must be one of name or mbid) and return the unique list. recordings also allow
        msid as key and artists allow artist_credit_id as key. Currently the order of the
        list is not preserved. This should be improved on...

        :param key: Which key to use for making the list unique. Defaults to "mbid".
    """

    def __init__(self, key="mbid"):
        troi.Element.__init__(self)
        self.key = key

    @staticmethod
    def inputs():
        return []

    def read(self, entities_arg):
        entities = entities_arg[0]

        if not entities:
            return []

        if not is_homogeneous(entities):
            raise TypeError("entity list not homogenous")
        _check_key_for_set_op(entities, self.key)

        entity_dict = {}
        for e in entities:
            if isinstance(e, troi.ArtistCredit) and self.key == "artists":
                entity_dict[",".join(getattr(e, self.key))] = e
            else:
                entity_dict[getattr(e, self.key)] = e

        return list(entity_dict.values())


class UnionElement(troi.Element):
    """ Combine both entities lists into one """

    def __init__(self):
        troi.Element.__init__(self)

    def inputs(self):
        return []

    def read(self, entities):
        entities_0 = entities[0]
        entities_1 = entities[1]

        if not entities_0:
            return entities_1

        if not entities_1:
            return entities_0

        _ensure_conformity(entities_0, entities_1)

        result = copy.copy(entities_0)
        result.extend(entities_1)

        return result


class IntersectionElement(troi.Element):
    """
        Return the list of entities that exist in both entities lists.
    """

    def __init__(self, key="mbid"):
        troi.Element.__init__(self)
        self.key = key

    def inputs(self):
        return []

    def read(self, entities):
        entities_0 = entities[0]
        entities_1 = entities[1]

        if not entities_0 or not entities_1:
            return []

        _ensure_conformity(entities_0, entities_1)
        _check_key_for_set_op(entities_0, self.key)

        entity_dict = {}
        for e in entities_1:
            if isinstance(e, troi.Artist) and self.key == "mbids":
                entity_dict[",".join(getattr(e, self.key))] = e
            else:
                entity_dict[getattr(e, self.key)] = e

        results = []
        for e in entities_0:
            if isinstance(e, troi.Artist) and self.key == "mbids":
                if ",".join(getattr(e, self.key)) in entity_dict:
                    results.append(e)
            else:
                if getattr(e, self.key) in entity_dict:
                    results.append(e)


        return results


class DifferenceElement(troi.Element):
    '''
        Return the list of recordings in entities_0 minus those in entities_1
    '''

    def __init__(self, key = "mbid"):
        troi.Element.__init__(self)
        self.key = key

    def inputs(self):
        return []

    def read(self, entities):
        entities_0 = entities[0]
        entities_1 = entities[1]

        if not entities_0:
            return []

        if not entities_1:
            return entities_0

        _ensure_conformity(entities_0, entities_1)
        _check_key_for_set_op(entities_0, self.key)

        entity_dict = {}
        for e in entities_1:
            if isinstance(e, troi.Artist) and self.key == "mbids":
                entity_dict[",".join(getattr(e, self.key))] = e
            else:
                entity_dict[getattr(e, self.key)] = e

        results = []
        for e in entities_0:
            if isinstance(e, troi.Artist) and self.key == "mbids":
                if ",".join(getattr(e, self.key)) not in entity_dict:
                    results.append(e)
            else:
                if getattr(e, self.key) not in entity_dict:
                    results.append(e)

        return results


class ZipperElement(troi.Element):
    '''
        Given two or more inputs, pick recordings from each alternatingly
    '''

    def __init__(self):
        troi.Element.__init__(self)

    @staticmethod
    def inputs():
        return [troi.Recording, troi.Recording]

    @staticmethod
    def outputs():
        return [troi.Recording]

    def read(self, inputs):
        output = []
        for rec0, rec1 in zip_longest(inputs[0], inputs[1]):
            if rec0 is not None:
                output.append(rec0)
            if rec1 is not None:
                output.append(rec1)

        return output



from collections import defaultdict
from datetime import datetime, timedelta
from typing import Optional
from time import sleep

import requests

from troi import Element, Recording


class ListensFeedbackLookup(Element):
    """ Element to look up the user's feedback for the given Recordings.

        :param user_name: The ListenBrainz user_name for whom to fetch feedback information.
    """

    def __init__(self, user_name, auth_token=None):
        super().__init__()
        self.user_name = user_name
        self.auth_token = auth_token

    @staticmethod
    def inputs():
        return [Recording]

    @staticmethod
    def outputs():
        return [Recording]

    def read(self, inputs):
        recordings = inputs[0]
        if not recordings:
            return []

        mbids = set()
        for r in recordings:
            mbids.add(r.mbid)
        mbids = list(mbids)

        headers = {"Authorization": f"Token {self.auth_token}"} if self.auth_token else {}

        feedback_map = {}
        batch_size = 50
        for idx in range(0, len(mbids), batch_size):
            recording_mbids = mbids[idx: idx + batch_size]

            while True:
                response = requests.get(
                    f"https://api.listenbrainz.org/1/feedback/user/{self.user_name}/get-feedback-for-recordings",
                    params={"recording_mbids": ",".join(recording_mbids)},
                    headers=headers
                )
                if response.status_code == 429:
                    sleep(2)
                    continue

                break
            response.raise_for_status()
            data = response.json()["feedback"]
            if len(data) == 0:
                continue

            for feedback in data:
                mbid = feedback["recording_mbid"]
                feedback_map[mbid] = feedback["score"]

        for r in recordings:
            r.listenbrainz["score"] = feedback_map.get(r.mbid, 0)

        return recordings



import datetime
import logging

from libsonic.errors import DataNotFoundError
import peewee
from tqdm import tqdm

from troi.content_resolver.database import Database
from troi.content_resolver.model.database import db
from troi.content_resolver.model.recording import Recording, FileIdType
from troi.content_resolver.utils import bcolors
from troi.content_resolver.py_sonic_fix import FixedConnection

logger = logging.getLogger(__name__)


class SubsonicDatabase(Database):
    '''
    Add subsonic sync capabilities to the Database
    '''

    # Determined by the number of albums we can fetch in one go
    BATCH_SIZE = 500

    def __init__(self, index_dir, config, quiet):
        self.config = config
        Database.__init__(self, index_dir, quiet)
        self.quiet = quiet

    def sync(self):
        """
            Scan the subsonic client specified in config.py
        """

        # Keep some stats
        self.total = 0
        self.matched = 0
        self.error = 0

        self.run_sync()
        
        logger.info("Checked %s albums:" % self.total)
        logger.info("  %5d albums matched" % self.matched)
        logger.info("  %5d recordings with errors" % self.error)

    def connect(self):
        if not self.config:
            logger.error("Missing credentials to connect to subsonic")
            return None

        logger.info("[ connect to subsonic ]")

        return FixedConnection(
            self.config.SUBSONIC_HOST,
            self.config.SUBSONIC_USER,
            self.config.SUBSONIC_PASSWORD,
            self.config.SUBSONIC_PORT,
        )

    def run_sync(self):
        """
            Perform the sync between the local collection and the subsonic one.
        """

        conn = self.connect()
        if not conn:
            return

        cursor = db.connection().cursor()

        logger.info("[ load albums ]")
        album_ids = set()
        albums = []
        offset = 0
        while True:
            results = conn.getAlbumList2(ltype="alphabeticalByArtist", size=self.BATCH_SIZE, offset=offset)
            albums.extend(results["albumList2"]["album"])
            album_ids.update([r["id"] for r in results["albumList2"]["album"]])

            album_count = len(results["albumList2"]["album"])
            offset += album_count
            if album_count < self.BATCH_SIZE:
                break

        logger.info("[ loaded %d albums ]" % len(album_ids))

        if not self.quiet:
            pbar = tqdm(total=len(album_ids))
        recordings = []

        # cross reference subsonic artist id to artitst_mbid
        artist_id_index = {}

        for album in albums:
            album_info = conn.getAlbum(id=album["id"])

            # Some servers might already include the MBID in the list or album response
            album_mbid = album_info.get("musicBrainzId", album.get("musicBrainzId"))
            if not album_mbid:
                album_info2 = conn.getAlbumInfo2(aid=album["id"])
                try:
                    album_mbid = album_info2["albumInfo"]["musicBrainzId"]
                except KeyError:
                    if not self.quiet:
                        pbar.write(bcolors.FAIL + "FAIL " + bcolors.ENDC + "subsonic album '%s' by '%s' has no MBID" %
                                   (album["name"], album["artist"]))
                    self.error += 1
                    continue


            recordings = []
            for song in album_info["album"]["song"]:
                album = album_info["album"]
            
                if "artistId" in song:
                    artist_id = song["artistId"]
                else:
                    artist_id = album["artistId"]

                if artist_id not in artist_id_index:
                    artist = conn.getArtistInfo2(artist_id)
                    try:
                        artist_id_index[artist_id] = artist["artistInfo2"]["musicBrainzId"]
                    except KeyError:
                        if not self.quiet:
                            pbar.write(bcolors.FAIL + "FAIL " + bcolors.ENDC + "recording '%s' by '%s' has no artist MBID" %
                                    (album["name"], album["artist"]))
                            pbar.write("Consider retagging this file with Picard! ( https://picard.musicbrainz.org )")
                        self.error += 1
                        continue

                self.add_subsonic({
                    "artist_name": song["artist"],
                    "release_name": song["album"],
                    "recording_name": song["title"],
                    "artist_mbid": artist_id_index[artist_id],
                    "release_mbid": album_mbid,
                    "recording_mbid": song["musicBrainzId"],
                    "duration": song["duration"] * 1000,
                    # Neither track number nor disc number are guaranteed for subsonic
                    "track_num": song.get("track", 1),
                    "disc_num": song.get("discNumber", 1),
                    "subsonic_id": song["id"],
                    "mtime": datetime.datetime.now()
                    })

            if not self.quiet:
                pbar.write(bcolors.OKGREEN + "OK   " + bcolors.ENDC + "album %-50s %-50s" %
                           (album["name"][:49], album["artist"][:49]))
            self.matched += 1
            self.total += 1
            if not self.quiet:
                pbar.update(1)

        if len(recordings) >= self.BATCH_SIZE:
            self.update_recordings(recordings)

    def add_subsonic(self, mdata):
        """ 
            Given recording metadata, add it to the database or update it if it already exists
            update the recording instead
        """

        with db.atomic() as transaction:
            try:
                recording = Recording.select().where(Recording.file_id == mdata['subsonic_id']).get()
                recording.artist_name = mdata["artist_name"]
                recording.release_name = mdata["release_name"]
                recording.recording_name = mdata["recording_name"]
                recording.artist_mbid = mdata["artist_mbid"]
                recording.release_mbid = mdata["release_mbid"]
                recording.recording_mbid = mdata["recording_mbid"]
                recording.mtime = mdata["mtime"]
                recording.track_num = mdata["track_num"]
                recording.disc_num = mdata["disc_num"]
                recording.save()
                return "updated"
            except peewee.DoesNotExist:
                recording = Recording.create(file_id=mdata["subsonic_id"],
                                             file_id_type=FileIdType(FileIdType.SUBSONIC_ID),
                                             artist_name=mdata["artist_name"],
                                             release_name=mdata["release_name"],
                                             recording_name=mdata["recording_name"],
                                             artist_mbid=mdata["artist_mbid"],
                                             release_mbid=mdata["release_mbid"],
                                             recording_mbid=mdata["recording_mbid"],
                                             mtime=mdata["mtime"],
                                             duration=mdata["duration"],
                                             track_num=mdata["track_num"],
                                             disc_num=mdata["disc_num"])
                recording.save()
                return "added"


    def update_recordings(self, recordings):
        """
            Given a list of recording_subsonic records, update the DB.
            Updates recording_id, subsonic_id, last_update
        """

        recording_index = { r[0]:r[1] for r in recordings }

        cursor = db.connection().cursor()
        with db.atomic():

            placeholders = ",".join(("?", ) * len(recording_index))
            cursor.execute("""SELECT recording_id
                                FROM recording_subsonic
                               WHERE recording_id in (%s)""" % placeholders, tuple(recording_index.keys()))
            existing_ids = { row[0]:None for row in cursor.fetchall() }
            existing_recordings = []
            new_recordings = []
            for r in recordings:
                if r[0] in existing_ids:
                    existing_recordings.append((r[0], r[1], datetime.datetime.now(), r[0]))
                else:
                    new_recordings.append((r[0], r[1], datetime.datetime.now()))

            cursor.executemany("""INSERT INTO recording_subsonic (recording_id, subsonic_id, last_updated)
                                       VALUES (?, ?, ?)""", tuple(new_recordings))

            cursor.executemany("""UPDATE recording_subsonic
                                     SET recording_id = ?
                                       , subsonic_id = ?
                                       , last_updated = ?
                                   WHERE recording_id = ?""", tuple(existing_recordings))


        # This concise query does the same as above. But older versions of python/sqlite on Raspberry Pis 
        # don't support upserts yet. :(
        #recordings = [(r[0], r[1], datetime.datetime.now()) for r in recordings]
        #cursor.executemany(
        #    """INSERT INTO recording_subsonic (recording_id, subsonic_id, last_updated)
        #                            VALUES (?, ?, ?)
        #         ON CONFLICT DO UPDATE SET recording_id = excluded.recording_id
        #                                 , subsonic_id = excluded.subsonic_id
        #                                 , last_updated = excluded.last_updated""", recordings)

    def upload_playlist(self, playlist, playlist_id=None):
        """
            Given a Troi playlist, upload the playlist to the subsonic API.
        """

        conn = self.connect()
        if not conn:
            return

        song_ids = []
        for recording in playlist.playlists[0].recordings:
            try:
                song_ids.append(recording.musicbrainz["subsonic_id"])
            except KeyError:
                continue

        if playlist_id:
            try:
                remote_playlist = conn.getPlaylist(pid=playlist_id)
                removed_song_idx = list(range(0, remote_playlist["playlist"]["songCount"]))
                conn.updatePlaylist(
                    lid=playlist_id,
                    name=playlist.playlists[0].name,
                    songIdsToAdd=song_ids,
                    songIndexesToRemove=removed_song_idx,
                )
            except DataNotFoundError:
                conn.createPlaylist(name=playlist.playlists[0].name, songIds=song_ids)
        else:
            conn.createPlaylist(name=playlist.playlists[0].name, songIds=song_ids)



import os
import datetime
import logging
from math import fabs
from time import time
import re
import sys

from sklearn.feature_extraction.text import TfidfVectorizer
from unidecode import unidecode

try:
    import nmslib
    have_nmslib = True
except ImportError:
    have_nmslib = False

logger = logging.getLogger(__name__)

def ngrams(string, n=3):
    """ Take a lookup string (noise removed, lower case, etc) and turn into a list of trigrams """

    string = ' ' + string + ' '  # pad names for ngrams...
    ngrams = zip(*[string[i:] for i in range(n)])
    return [''.join(ngram) for ngram in ngrams]


class FuzzyIndex:
    '''
       Create a fuzzy index using a Term Frequency, Inverse Document Frequency (tf-idf)
       algorithm. Currently the libraries that implement this cannot be serialized to disk,
       so this is an in memory operation. Fortunately for our amounts of data, it should
       be quick to rebuild this index.
    '''

    def __init__(self):
        global have_nmslib

        self.have_nmslib = have_nmslib
        self.vectorizer = None
        self.index = None

    def encode_string(self, text):
        if text is None:
            return None
        return unidecode(re.sub(" +", "", re.sub(r'[^\w ]+', '', text)).strip().lower())

    def build(self, artist_recording_data):
        """
            Builds a new index and saves it to disk and keeps it in ram as well.
        """

        if not self.have_nmslib:
            return

        self.lookup_strings = []
        lookup_ids = []
        for artist_name, recording_name, lookup_id in artist_recording_data:
            if artist_name is None or recording_name is None:
                continue
            self.lookup_strings.append(self.encode_string(artist_name) + self.encode_string(recording_name))
            lookup_ids.append(lookup_id)

        self.vectorizer = TfidfVectorizer(min_df=1, analyzer=ngrams)
        lookup_matrix = self.vectorizer.fit_transform(self.lookup_strings)

        self.index = nmslib.init(method='simple_invindx', space='negdotprod_sparse_fast', data_type=nmslib.DataType.SPARSE_VECTOR)
        self.index.addDataPointBatch(lookup_matrix, lookup_ids)
        self.index.createIndex()

    def search(self, query_data):
        """
            Return IDs for the matches in a list. Returns a list of dicts with keys of lookup_string, confidence and recording_id.
        """
        if not self.have_nmslib:
            logger.warn("nmslib not installed and trying fuzzy search, but nothing will match. Install nmslib!")
            return []

        query_strings = []
        for data in query_data:
            if data["artist_name"] is None or data["recording_name"] is None:
                continue

            query_strings.append(self.encode_string(data["artist_name"]) + self.encode_string(data["recording_name"]))

        query_matrix = self.vectorizer.transform(query_strings)
        results = self.index.knnQueryBatch(query_matrix, k=1, num_threads=1)

        output = []
        for i, result in enumerate(results):
            if len(result[0]):
                output.append({"confidence": fabs(result[1][0]),
                               "recording_id": result[0][0]})
            else:
                output.append({"confidence": 0.0, "recording_id": 0})

        return output



import json
import unittest
import unittest.mock

import troi
import troi.tools.area_lookup

request_json = [
    {
        "[area]": "Germany"
    }
]

return_json = [
    {
        "area_mbid": "85752fda-13c4-31a3-bee5-0e5cb1f51dad",
        "area_name": "Germany"
    }
]

class TestAreaLookup(unittest.TestCase):

    @unittest.mock.patch('requests.post')
    def test_area_lookup(self, req):

        mock = unittest.mock.MagicMock()
        mock.status_code = 200
        mock.text = json.dumps(return_json)
        req.return_value = mock
        area_id = troi.tools.area_lookup.area_lookup(request_json[0]["[area]"])
        req.assert_called_with(troi.tools.area_lookup.AREA_LOOKUP_SERVER_URL, json=request_json)

        assert area_id == "85752fda-13c4-31a3-bee5-0e5cb1f51dad"



import logging
import requests

from .utils import AppleMusicAPI, AppleMusicException
from more_itertools import chunked

logger = logging.getLogger(__name__)

APPLE_MUSIC_URL = f"https://api.music.apple.com/"
APPLE_MUSIC_IDS_LOOKUP_URL = "https://labs.api.listenbrainz.org/apple-music-id-from-mbid/json"

def lookup_apple_music_ids(recordings):
    """ Given a list of Recording elements, try to find Apple Music track ids from labs api apple_music_id lookup_from_mbid
    and add those to the recordings. """
    response = requests.post(
        APPLE_MUSIC_IDS_LOOKUP_URL,
        json=[{"recording_mbid": recording.mbid} for recording in recordings]
    )
    response.raise_for_status()
    apple_data = response.json()

    mbid_apple_music_ids_index = {}
    apple_id_mbid_index = {}
    for recording, lookup in zip(recordings, apple_data):
        if len(lookup["apple_music_track_ids"]) > 0:
            recording.apple_music_id = lookup["apple_music_track_ids"][0]
            mbid_apple_music_ids_index[recording.mbid] = lookup["apple_music_track_ids"]
            for apple_music_id in lookup["apple_music_track_ids"]:
                apple_id_mbid_index[apple_music_id] = recording.mbid
    return recordings, mbid_apple_music_ids_index, apple_id_mbid_index

def get_tracks_from_apple_playlist(developer_token, user_token, playlist_id):
    """ Get tracks from the Apple Music playlist.
    """
    apple = AppleMusicAPI(developer_token, user_token)
    tracks, name, description = apple.get_playlist_tracks(playlist_id)

    mapped_tracks = [
        {
            "recording_name": track['attributes']['name'],
            "artist_name": track['attributes']['artistName']
        }
        for track in tracks
    ]

    return mapped_tracks, name, description

def submit_to_apple_music(apple: AppleMusicAPI, playlist, is_public: bool=True, existing_url=None):
    """ Submit or update an existing Apple Music playlist.

    If existing urls are specified then isPublic arguments are ignored.
    """
    filtered_recordings = [recording for recording in playlist.recordings if recording.mbid]
    _, mbid_apple_music_index, apple_music_mbid_index = lookup_apple_music_ids(filtered_recordings)
    apple_music_track_ids = [recording.apple_music_id for recording in filtered_recordings if recording.apple_music_id]
    if len(apple_music_track_ids) == 0:
        return None, None

    logger.info("submit %d tracks" % len(apple_music_track_ids))

    playlist_id, playlist_url = None, None

    if existing_url:
        logger.info("Apple Music API does not support updating playlists, --apple-music-url paramenter will be ignored. Creating a new playlist...")

    if not playlist_id:
        # create new playlist
        apple_playlist = apple.create_playlist(
            name=playlist.name,
            description=playlist.description,
            is_public=is_public
        )
        playlist_id = apple_playlist["data"][0]["id"]
        playlist_url = apple_playlist["data"][0]["href"]

    for chunk in chunked(apple_music_track_ids, 100):
        apple.playlist_add_tracks(playlist_id, chunk)

    playlist.add_metadata({"external_urls": {"apple_music": playlist_url}})

    return playlist_url, playlist_id



import requests
import logging

from collections import defaultdict
from more_itertools import chunked
from .utils import SoundcloudAPI, SoundCloudException

logger = logging.getLogger(__name__)

SOUNDCLOUD_IDS_LOOKUP_URL = "https://labs.api.listenbrainz.org/soundcloud-id-from-mbid/json"

def lookup_soundcloud_ids(recordings):
    """ Given a list of Recording elements, try to find soundcloud track ids from labs api soundcloud lookup using mbids
    and add those to the recordings. """
    response = requests.post(
        SOUNDCLOUD_IDS_LOOKUP_URL,
        json=[{"recording_mbid": recording.mbid} for recording in recordings]
    )
    response.raise_for_status()

    soundcloud_data = response.json()
    mbid_soundcloud_ids_index = {}
    soundcloud_id_mbid_index = {}
    for recording, lookup in zip(recordings, soundcloud_data):
        if len(lookup["soundcloud_track_ids"]) > 0:
            recording.soundcloud_id = lookup["soundcloud_track_ids"][0]
            mbid_soundcloud_ids_index[recording.mbid] = lookup["soundcloud_track_ids"]
            for soundcloud_id in lookup["soundcloud_track_ids"]:
                soundcloud_id_mbid_index[soundcloud_id] = recording.mbid
    return recordings, mbid_soundcloud_ids_index, soundcloud_id_mbid_index


def _check_unplayable_tracks(soundcloud: SoundcloudAPI, playlist_id: str):
    """ Retrieve tracks for given soundcloud playlist and split into lists of playable and unplayable tracks """
    tracks = soundcloud.get_playlist_tracks(playlist_id, linked_partitioning=True, limit=100, access=['playable, preview ,blocked'])
    track_details = [
        {
            "id": track["id"],
            "title": track["title"],
            "access": track["access"]
        }
        for track in tracks
    ]

    playable = []
    unplayable = []
    for idx, item in enumerate(track_details):
        if item["access"] == "playable":
            playable.append((idx, item["id"]))
        else:
            unplayable.append((idx, item["id"]))
    return playable, unplayable


def _get_alternative_track_ids(unplayable, mbid_soundcloud_id_idx, soundcloud_id_mbid_idx):
    """ For the list of unplayable track ids, find alternative tracks ids.

    mbid_soundcloud_id_idx is an index with mbid as key and list of equivalent soundcloud ids as value.
    soundcloud_id_mbid_idx is an index with soundcloud_id as key and the corresponding mbid as value.
    """
    index = defaultdict(list)
    soundcloud_ids = []
    for idx, soundcloud_id in unplayable:
        mbid = soundcloud_id_mbid_idx[str(soundcloud_id)]
        other_soundcloud_ids = mbid_soundcloud_id_idx[mbid]

        for new_idx, new_soundcloud_id in enumerate(other_soundcloud_ids):
            if new_soundcloud_id == soundcloud_id:
                continue
            soundcloud_ids.append(new_soundcloud_id)
            index[idx].append(new_soundcloud_id)
            
    return soundcloud_ids, index


def _get_fixed_up_tracks(soundcloud: SoundcloudAPI, soundcloud_ids, index):
    """ Lookup the all alternative soundcloud track ids, filter playable ones and if multiple track ids
        for same item match, prefer the one occurring earlier. If no alternative is playable, ignore the
        item altogether.
    """
    new_tracks = soundcloud.get_track_details(soundcloud_ids)

    new_tracks_ids = set()
    for item in new_tracks:
        if item["access"] == "playable":
            new_tracks_ids.add(item["id"])

    fixed_up_items = []
    for idx, soundcloud_ids in index.items():
        for soundcloud_id in soundcloud_ids:
            if soundcloud_id in new_tracks_ids:
                fixed_up_items.append((idx, soundcloud_id))
                break
    return fixed_up_items


def fixup_soundcloud_playlist(soundcloud: SoundcloudAPI, playlist_id: str, mbid_soundcloud_id_idx, soundcloud_id_mbid_idx):
    """ Fix unplayable tracks in the given soundcloud playlist.

    Given a soundcloud playlist id, look it up and find unstreamable tracks. If there are any unstreamable tracks, try
    alternative soundcloud track ids from index/reverse_index if available. If alternative is not available or alternatives
    also are unplayable, remove the track entirely from the playlist. Finally, update the playlist if needed.
    """
    playable, unplayable = _check_unplayable_tracks(soundcloud, playlist_id)
    if not unplayable:
        return

    alternative_ids, index = _get_alternative_track_ids(unplayable, mbid_soundcloud_id_idx, soundcloud_id_mbid_idx)
    if not alternative_ids:
        return

    fixed_up = _get_fixed_up_tracks(soundcloud, alternative_ids, index)
    all_items = []
    all_items.extend(playable)
    all_items.extend(fixed_up)

    # sort all items by index value to ensure the order of tracks of original playlist is preserved
    all_items.sort(key=lambda x: x[0])
    # update all track ids the soundcloud playlist
    finalized_ids = [x[1] for x in all_items]

    # clear existing playlist
    soundcloud.update_playlist_details(playlist_id, track_ids=[])

    # chunking requests to avoid hitting rate limits
    for chunk in chunked(finalized_ids, 100):
        soundcloud.add_playlist_tracks(playlist_id, chunk)


def get_tracks_from_soundcloud_playlist(developer_token, playlist_id):
    """ Get tracks from the Soundcloud playlist.
    """
    soundcloud = SoundcloudAPI(developer_token)
    tracks = soundcloud.get_playlist_tracks(playlist_id, linked_partitioning=True, limit=100, access=['playable, preview ,blocked'])

    mapped_tracks = [
        {
            "recording_name": track['title'].split(" - ")[1] if " - " in track['title'] else track['title'],
            "artist_name": track['title'].split(" - ")[0] if " - " in track['title'] else track['user']['username']
        }
        for track in tracks
    ]

    return mapped_tracks


def get_soundcloud_playlist(developer_token, playlist_id):
    soundcloud = SoundcloudAPI(developer_token)
    data = soundcloud.get_playlist(playlist_id)
    name = data["title"]
    description = data["description"]

    if description is None:
        description = ""

    return [], name, description


def submit_to_soundcloud(soundcloud: SoundcloudAPI, playlist, is_public: bool = True,
                    existing_url: str = None):
    """ Submit or update an existing soundcloud playlist.

    If existing urls are specified then is_public and is_collaborative arguments are ignored.
    """
    filtered_recordings = [r for r in playlist.recordings if r.mbid]

    _, mbid_soundcloud_index, soundcloud_mbid_index = lookup_soundcloud_ids(filtered_recordings)
    soundcloud_track_ids = [r.soundcloud_id for r in filtered_recordings if r.soundcloud_id]
    if len(soundcloud_track_ids) == 0:
        return None, None

    logger.info("submit %d tracks" % len(soundcloud_track_ids))

    playlist_id, playlist_url = None, None
    if existing_url:
        # update existing playlist
        playlist_url = existing_url
        playlist_id = existing_url.split("/")[-1]
        try:
            soundcloud.update_playlist_details(playlist_id=playlist_id, title=playlist.name, description=playlist.description)
        except SoundCloudException as err:
            # one possibility is that the user has deleted the soundcloud from playlist, so try creating a new one
            logger.info("provided playlist url has been unfollowed/deleted by the user, creating a new one")
            playlist_id, playlist_url = None, None

    if not playlist_id:
        # create new playlist
        soundcloud_playlist = soundcloud.create_playlist(
            title=playlist.name,
            sharing=is_public,
            description=playlist.description
        )
        playlist_id = soundcloud_playlist["id"]
        playlist_url = soundcloud_playlist["permalink"]
    else:
        # existing playlist, clear it
        tracks = map(lambda id: dict(id=id), [0])
        soundcloud.update_playlist(playlist_id, tracks)

    for chunk in chunked(soundcloud_track_ids, 100):
        soundcloud.add_playlist_tracks(playlist_id, chunk)

    fixup_soundcloud_playlist(soundcloud, playlist_id, mbid_soundcloud_index, soundcloud_mbid_index)

    playlist.add_metadata({"external_urls": {"soundcloud": playlist_url}})

    return playlist_url, playlist_id


from peewee import SqliteDatabase

PRAGMAS = (
    ('foreign_keys', 1),
    ('journal_mode', 'WAL'),
)

db = SqliteDatabase(None, pragmas=PRAGMAS)


def setup_db(db_file):
    global db
    db.init(db_file)



import mutagen.asf

from troi.content_resolver.formats.tag_utils import get_tag_value, extract_track_number


EXTENSIONS = {'.wma'}
READER = mutagen.asf.ASF


def get_metadata(tags):

    mdata = {}
    mdata["artist_name"] = str(get_tag_value(tags, "Author"))
    mdata["artist_sortname"] = str(get_tag_value(tags, "WM/ArtistSortOrde", mdata["artist_name"]))
    mdata["release_name"] = str(get_tag_value(tags, "WM/AlbumTitle"))
    mdata["recording_name"] = str(get_tag_value(tags, "Title"))
    mdata["track_num"] = extract_track_number(str(get_tag_value(tags, "WM/TrackNumber")))
    mdata["disc_num"] = int(get_tag_value(tags, "WM/SetSubTitle") or 1)
    mdata["artist_mbid"] = str(get_tag_value(tags, "MusicBrainz/Artist Id"))
    mdata["recording_mbid"] = str(get_tag_value(tags, "MusicBrainz/Release Track Id"))
    mdata["release_mbid"] = str(get_tag_value(tags, "MusicBrainz/Album Id"))
    mdata["duration"] = int(tags.info.length * 1000)

    return mdata
